from typing import Any, Self
import torch.nn as ptn
from ...types import Tensor, Module, Functional


class GatedSumMixer(Module):
    """Combine stacked feature vectors by a per-instance linear combination.

    The per-instance coefficients sum to 1 for each data point and can thus be
    seen as some sort of per-instance feature importance. They are obtained by
    concatenating all features into a single, wide vector, and linearly
    projecting it down to a vector with twice as many elements as there are
    features. One half is then passed through an (optional) activation function
    to gate the other half, thus reducing the final output back down to the
    number of features to combine.

    Parameters
    ----------
    mod_dim: int
        Size of the feature space. The input tensor is expected to be of that
        size in its last dimension and the output will again have this size in
        its last dimension.
    n_features: int
        The number of features to combine. Must be equal to the size of the
        next-to-last dimension of the input tensor.
    gate: Module or function, optional
        The activation function to be applied to half of the (linearly)
        transformed inputs before multiplying with the other half. Must be
        a callable that accepts a tensor as sole argument, like a module from
        ``torch.nn`` or a function from ``torch.nn.functional``.
        Defaults to a sigmoid.
    **kwargs
        Keyword arguments are passed on to the linear layer.

    """

    def __init__(
            self,
            mod_dim: int,
            n_features: int,
            gate: Module | Functional = ptn.Sigmoid(),
            **kwargs: Any
    ) -> None:
        super().__init__()
        self.mod_dim = mod_dim
        self.n_features = n_features
        self.gate = gate
        self.kwargs = kwargs
        self.coeffs = ptn.Linear(
            n_features * mod_dim,
            2 * n_features, **kwargs
        )
        self.norm = ptn.Softmax(dim=-1)

    def importance(self, inp: Tensor) -> Tensor:
        """Per-instance weights in the normed linear combination of features.

        Parameters
        ----------
        inp: Tensor
            Feature vectors stacked into a tensor of at least 2 dimensions.
            The size of the next-to-last last dimension is expected to match
            the `n_features` provided at instantiation. The last dimension
            (of size `mod_dim`) is expected to contain the features vectors.

        Returns
        -------
        Tensor
            The output tensor has one fewer dimensions than the input with the
            last dimension being dropped.

        """
        mixed = self.coeffs(inp.flatten(start_dim=-2))
        gated = self.gate(mixed[..., self.n_features:])
        return self.norm(mixed[..., :self.n_features] * gated)

    def forward(self, inp: Tensor) -> Tensor:
        """Forward pass for combining multiple stacked feature vectors.

        Parameters
        ----------
        inp: Tensor
            Feature vectors stacked into a tensor of at least 2 dimensions.
            The size of the next-to-last last dimension is expected to match
            the `n_features` provided at instantiation. The last dimension
            (of size `mod_dim`) is expected to contain the features vectors.

        Returns
        -------
        Tensor
            The output tensor has one fewer dimensions than the input.
            The next-to-last dimension is dropped and the last dimension now
            contains the per-instance (normed) linear combination of all
            feature vectors.

        """
        return (self.importance(inp).unsqueeze(dim=-2) @ inp).squeeze(dim=-2)

    def reset_parameters(self) -> None:
        """Re-initialize all internal parameters."""
        self.coeffs.reset_parameters()

    def new(
            self,
            mod_dim: int | None = None,
            n_features: int | None = None,
            gate: Module | Functional | None = None,
            **kwargs: Any
    ) -> Self:
        """Return a fresh instance with the same or updated parameters.

        Parameters
        ----------
        mod_dim: int, optional
            Size of the feature space. The input tensor is expected to be of
            that size in its last dimension and the output will again have this
            size in its last dimension. Overwrites the `mod_dim` of the current
            instance if given. Defaults to ``None``.
        n_features: int, optional
            The number of features to combine. Must be equal to the size of the
            next-to-last dimension of the input tensor. Overwrites `n_features`
            of the current instance if given. Defaults to ``None``.
        gate: Module or function, optional
            The activation function to be applied to half of the (linearly)
            transformed input before multiplying with the other half. Must be
            a callable that accepts a tensor as sole argument, like a module
            from ``torch.nn`` or a function from ``torch.nn.functional``.
            Overwrites the `gate` of the current instance if given.
            Defaults to ``None``.
        **kwargs
            Additional keyword arguments are merged into the keyword arguments
            of the current instance and are then passed through to the linear
            layer together.

        Returns
        -------
        GatedSumMixer
            A fresh, new instance of itself.

        """
        return self.__class__(
            self.mod_dim if mod_dim is None else mod_dim,
            self.n_features if n_features is None else n_features,
            self.gate if gate is None else gate,
            **(self.kwargs | kwargs)
        )



from typing import Any, Self
from collections.abc import Iterator, Callable, Iterable
from functools import singledispatchmethod
from ..misc import IndentRepr
from .exceptions import PipeError

type Call = type | Callable[..., Any]


class Pipe[**P, T](IndentRepr):
    """Chain any number of callable objects into a single callable object.

    Arguments passed to the functional composition will be forwarded to the
    first callable in the chain. Subsequent callables will be called with
    the return value(s) of the previous callable in the chain. The return
    value of the functional composition is the return value of the last
    callable in the chain.

    Parameters
    ----------
    call: callable or iterable of callables, optional
        One callable or an iterator of callables to chain one after another.
        Defaults to an empty tuple.
    *calls: callable
        Additional callables to chain one after another.

    Raises
    ------
    PipeError
        If (any of) `call` or any of `calls` are not, in fact, callable.

    Notes
    -----
    Upon instantiation, the generic class can be type-annotated with the list
    of argument types of the first callable in the chain, followed by the
    return type of the last callable.

    """

    def __init__(self, call: Call | Iterable[Call] = (), *calls: Call) -> None:
        self.calls = self.__valid(call) + self.__valid(calls)
        super().__init__(self.calls)

    def __iter__(self) -> Iterator[Call]:
        # We could also iterate over instances of self ...
        return self.calls.__iter__()

    def __len__(self) -> int:
        return self.calls.__len__()

    def __bool__(self) -> bool:
        return bool(self.calls)

    def __contains__(self, item: Call) -> bool:
        return item in self.calls

    def __reversed__(self) -> NotImplemented:
        return NotImplemented

    @singledispatchmethod
    def __getitem__(self, index: int) -> Call:
        # We could also return instances of self ...
        return self.calls[index]

    @__getitem__.register
    def _(self, index: slice) -> Self:
        return self.__class__(*self.calls[index])

    def __eq__(self, other: Self) -> bool:
        if isinstance(other, self.__class__):
            return self.calls == other.calls
        return NotImplemented

    def __ne__(self, other: Self) -> bool:
        if isinstance(other, self.__class__):
            return self.calls != other.calls
        return NotImplemented

    def __add__(self, other: Call | Iterable[Call] | Self) -> Self:
        if isinstance(other, self.__class__):
            return self.__class__(self.calls, *other.calls)
        try:
            return self.__class__(self.calls, *self.__valid(other))
        except PipeError:
            return NotImplemented

    def __radd__(self, other: Call | Iterable[Call] | Self) -> Self:
        if isinstance(other, self.__class__):
            return self.__class__(other.calls, *self.calls)
        try:
            return self.__class__(self.__valid(other), *self.calls)
        except PipeError:
            return NotImplemented

    def __call__(self, *args: P.args) -> T:
        """Call the functional composition this object was instantiated with.

        Parameters
        ----------
        *args
            Arguments to pass to the first callable in `calls`.

        Returns
        -------
        object
            Whatever the last callable of `calls` returns.

        Raises
        ------
        PipeError
            When one of the callables in the chain raises an exception.

        """
        args = args[0] if len(args) == 1 else args
        for i, call in enumerate(self):
            try:
                args = call(*args) if isinstance(args, tuple) else call(args)
            except Exception as error:
                msg = '\n{} executing\n{}\nin step {} of\n{}\n{}'
                err_cls = error.__class__.__name__
                name = self._name(call)
                fmt = msg.format(err_cls, name, i, self, error)
                raise PipeError(fmt) from error
        return args

    @staticmethod
    def __valid(calls: Call | Iterable[Call]) -> tuple[Call, ...]:
        """Ensure that the argument is indeed an iterable of callables."""
        if callable(calls):
            return calls,
        iterable = True
        all_callable = False
        try:
            all_callable = all(callable(call) for call in calls)
        except TypeError:
            iterable = False
        if iterable and all_callable:
            return tuple(calls)
        raise PipeError('All items in the pipe must be callable!')



from typing import Any
from collections.abc import Hashable, Iterable, Callable
from numpy import dtype, ndarray
from pandas.core.dtypes.base import ExtensionDtype
from pandas import DataFrame, Series
from ..misc.repr import ReprName
from ..misc import ArgRepr

type Type = str | type | dtype | ExtensionDtype
type Types = dict[Hashable, Type]
type Mask = list[bool] | Series | ndarray
type Condition = Callable[[DataFrame], Mask]
type Transform = dict[Hashable, Any] | Series | Callable[[Any], Any]


class AsType(ReprName):
    """Partial of the ``pandas.DataFrame.astype`` method.

    Parameters
    ----------
    types: type or dict
        Single type or dictionary of column names and types, specifying type
        conversion of entire DataFrame or specific columns, respectively
    **kwargs
        Keyword arguments are passed on to the ``astype`` method call after
        the `types` argument.

    """
    def __init__(self, types: Type | Types, **kwargs: Any) -> None:
        super().__init__()
        self.types = types
        self.kwargs = kwargs

    def __repr__(self) -> str:
        kwargs = (f'{k}={self._repr(v)}' for k, v in self.kwargs.items())
        kwargs = ', '.join(kwargs)
        try:
            args = (f'{k!r}: {self._name(self.types[k])}' for k in self.types)
            args = '{' + ', '.join(args) + '}'
        except TypeError:
            args = self._name(self.types)
        signature = ', '.join(filter(None, [args, kwargs]))
        return f'{self.__class__.__name__}({signature})'

    def __call__(self, df: DataFrame) -> DataFrame:
        """Cast entire dataframe or columns thereof to specified types

        Parameters
        ----------
        df: DataFrame
            Pandas dataframe to type-cast.

        Returns
        -------
        DataFrame
            Pandas dataframe cast to new type or with columns cast to
            new types.

        """
        return df.astype(self.types, **self.kwargs)


class ColumnSelector(ArgRepr):
    """Select a single column of a pandas dataframe as a pandas series.

    This is simply a partial for calling a dataframe's ``__getitem__``
    method with a single argument (using the square-brackets accessor).

    Parameters
    ----------
    col: hashable
        Single DataFrame column to select.

    """

    def __init__(self, col: Hashable) -> None:
        self.col = self.__valid(col)
        super().__init__(col)

    def __call__(self, df: DataFrame) -> Series:
        """

        Parameters
        ----------
        df: DataFrame
            Pandas dataframe to select column from.

        Returns
        -------
        Series
            The selected dataframe column.

        """
        return df[self.col]

    @staticmethod
    def __valid(col: Hashable) -> Hashable:
        _ = hash(col)
        return col


class ColumnsSelector(ArgRepr):
    """Select one or more columns of a pandas dataframe as dataframe.

    This is simply a partial for calling a dataframe's ``__getitem__``
    method with a list of arguments (using the square-brackets accessor).

    Parameters
    ----------
    col: Hashable, optional
        Column name or iterable thereof. Defaults to an empty tuple.
    *cols: Hashable
        Additional columns names.

    """

    def __init__(
            self,
            col: Hashable | Iterable[Hashable] = (),
            *cols: Hashable
    ) -> None:
        col = self.__valid(col)
        self.cols: tuple[Hashable, ...] = col + self.__valid(cols)
        super().__init__(*self.cols)

    def __call__(self, df: DataFrame) -> DataFrame:
        """Select the specified column(s) from a pandas DataFrame.

        Parameters
        ----------
        df: DataFrame
            Pandas dataframe to select column(s) from.

        Returns
        -------
        DataFrame
            The selected dataframe column(s).

        """
        return df[list(self.cols)]

    @staticmethod
    def __valid(cols: Hashable | Iterable[Hashable]) -> tuple[Hashable, ...]:
        """Ensure that the columns are indeed an iterable of hashables."""
        if isinstance(cols, str):
            return cols,
        try:
            _ = [hash(col) for col in cols]
        except TypeError:
            _ = hash(cols)
            return cols,
        return tuple(cols)


class ColumnMapper(ArgRepr):
    """Transform one column of a pandas dataframe into another.

    This is simply a partial of calling the ``map`` method on one column of a
    dataframe and assigning the result to the same or another column of the
    same dataframe.

    Parameters
    ----------
    src_col: Hashable
        Column to call the ``map`` method on.
    transform: callable, Mapping, or Series
        Function or mapping in the form of a dictionary or a pandas series.
    tgt_col: Hashable, optional
        Dataframe column to store the series resulting from the
        transformation. Defaults to `src_col`, thus overwriting it in place.
    **kwargs
        Keyword arguments are passed on to the call of the Series'
        ``map`` method.

    """

    def __init__(
            self,
            src_col: Hashable,
            transform: Transform,
            tgt_col: Hashable | None = None,
            **kwargs: Any
    ) -> None:
        self.src_col = src_col
        self.tgt_col = src_col if tgt_col is None else tgt_col
        self.transform = transform
        self.kwargs = kwargs
        name = transform if callable(transform) else type(transform)
        super().__init__(src_col, name, tgt_col, **kwargs)

    def __call__(self, df: DataFrame) -> DataFrame:
        """Called the ``map`` method on a specified column of a DataFrame.

        Cached keyword arguments are forwarded to the method call and
        the result is stored in the specified column of the DataFrame.

        Parameters
        ----------
        df: DataFrame
            Pandas dataframe with the column to call the ``map`` method on.

        Returns
        -------
        DataFrame
            Pandas dataframe with the result of the column transformation in
            the specified column.


        """
        df[self.tgt_col] = df[self.src_col].map(self.transform, **self.kwargs)
        return df


class RowsSelector(ArgRepr):
    """Select rows from a pandas dataframe with a boolean mask or function.

    This is simply a partial for calling a dataframe's ``__getitem__``
    method (using the square-brackets accessor) with a callable that takes
    the dataframe as input, and produces a 1-D, boolean array-like structure
    (of the same length as the dataframe to select from).

    Parameters
    ----------
    condition: callable or array-like
        A callable that accepts a dataframe and produces a 1-D, boolean array-
        like structure of the same length

    """

    def __init__(self, condition: Condition) -> None:
        super().__init__(condition)
        self.condition = condition

    def __call__(self, df: DataFrame) -> DataFrame:
        """Select rows from a dataframe with the specified mask or condition.

        Parameters
        ----------
        df: DataFrame
            The pandas dataframe to select rows from.

        Returns
        -------
        DataFrame
            The pandas dataframe with only the selected rows.

        """
        return df[self.condition]



from typing import Any, Self
import torch.nn as ptn
from ..types import Tensor, Module, Functional
from ..misc import identity


class ActivatedEmbedder(Module):
    """Simple linear projection of an individual feature into embedding space.

    Parameters
    ----------
    mod_dim: int
        Desired embedding size. Will become the size of the last dimension of
        the output tensor.
    activate: Module or function, optional
        The activation function to be applied after (linear) projection into
        embedding space. Must be a callable that accepts a tensor as sole
        argument, like a module from ``torch.nn`` or a function from
        `torch.nn.functional``, depending on whether it needs to be further
        parameterized or not. Defaults to ``identity``, resulting in no
        non-linear activation whatsoever.
    inp_dim: int, optional
        The number of features to embed together. Defaults to 1.
    **kwargs
        Additional keyword arguments to pass through to the linear layer.

    """

    def __init__(
            self,
            mod_dim: int,
            activate: Module | Functional = identity,
            inp_dim: int = 1,
            **kwargs: Any
    ) -> None:
        super().__init__()
        self.mod_dim = mod_dim
        self.activate = activate
        self.inp_dim = inp_dim
        self.kwargs = kwargs
        self.embed = ptn.Linear(inp_dim, mod_dim, **kwargs)

    def forward(self, inp: Tensor) -> Tensor:
        """Embed a single numerical feature through a (non-)linear projection.

        Parameters
        ----------
        inp: Tensor
            The last dimension of the input tensor is typically expected to be
            of size 1 and to contain the numerical value of a single feature.
            In case `inp_dim` dim was explicitly set to a value > 1 on
            instantiation, the size of the last dimension must match `inp_dim`,
            the number of numerical features to embed together.

        Returns
        -------
        Tensor
            The output has the same number of dimensions as the input with
            the size of the last dimension changed to the specified `mod_dim`.

        """
        return self.activate(self.embed(inp))

    def reset_parameters(self) -> None:
        """Re-initialize all internal parameters."""
        self.embed.reset_parameters()

    def new(
            self,
            mod_dim: int | None = None,
            activate: Module | Functional | None = None,
            inp_dim: int | None = None,
            **kwargs: Any
    ) -> Self:
        """Return a fresh instance with the same or updated parameters.

        Parameters
        ----------
        mod_dim: int, optional
            Desired embedding size. Will become the size of the last dimension
            of the output tensor. Overwrites the `mod_dim` of the current
            instance if given. Defaults to ``None``.
        activate: Module or function, optional
            The activation function to be applied after (linear) projection
            into embedding space. Must be a callable that accepts a tensor as
            sole argument, like a module from ``torch.nn`` or a function from
            `torch.nn.functional``, depending on whether it needs to be further
            parameterized or not. Overwrites the `activate` of the current
            instance if given. Defaults to ``None``.
        inp_dim: int, optional
            The number of features to embed together. Overwrites the `inp_dim`
            of the current instance if given. Defaults to ``None``.
        **kwargs
            Additional keyword arguments are merged into the keyword arguments
            of the current instance and are then passed through to the linear
            layer together.

        Returns
        -------
        ActivatedEmbedder
            A fresh, new instance of itself.

        """
        return self.__class__(
            self.mod_dim if mod_dim is None else mod_dim,
            self.activate if activate is None else activate,
            self.inp_dim if inp_dim is None else inp_dim,
            **(self.kwargs | kwargs)
        )



import torch as pt
from ...misc import ArgRepr
from ...funcflow import Curry
from ..types import Module, Optimizer, LRScheduler
from ..exceptions import TrainError
from .callbacks import EpochCallback, EpochPrinter, TrainCallback, TrainPrinter
from .checkpoints import Checkpoint, InMemory
from .schedulers import NoSchedule
from .data import TestDataBase, TrainDataBase

__all__ = ['Trainer']


class Trainer(ArgRepr):
    """Train and, optionally, evaluate a model with early stopping.

    Parameters
    ----------
    batch_size: int
        Size of the mini-batches to request from the training data.
    max_epochs: int
        Maximum number of epochs to train for.
    loss: Module
        PyTorch module that accepts the output(s) of the model to train as
        first argument(s) and the target as last argument and that produces
        the (scalar) loss to minimize.
    optimizer: Curry[Optimizer]
        A curry of a preconfigured PyTorch Optimizer (or some other custom
        construct) that returns a fully configured PyTorch Optimizer when
        called with the ``model.parameters()`` of the model to train.
    scheduler: Curry[LRScheduler], optional
        A curry of a PyTorch learning-rate scheduler (or some other custom
        construct) that returns a fully configured learning-rate scheduler
        when called with a PyTorch optimizer. Defaults to never changing the
        learning rate at all.
    warmup: int, optional
        Number of epochs to wait until the learning rate scheduler is used
        for the first time. Defaults to 0.
    patience: int, optional
        If patience is not ``None`` and smaller than `max_epochs`, early
        stopping is active. A snapshot of the model's state is taken after
        each epoch that improved the loss below its last minimum. If no
        improvement occurs for `patience` epochs, model training is stopped
        (even if `max_epochs` has not been reached yet) and the model is reset
        to its best state. If `warmup` > 0, then the early stopping gets active
        only after `warmup` epochs have passed.
    max_n: int, optional
        Maximum number of data points to take from the training (and,
        if present, test) data for computing the train (and, optional, test)
        loss after each epoch. Since this is done in a single batch, memory
        consumption might momentarily spike if not set or set too large.
        Defaults to number of data points in test set if present or train set
        if not.
    checkpoint: Checkpoint, optional
        Whenever the train (or test) loss after an epoch is smaller than the
        loss after the last, a new snapshot of the model state is saved by
        calling the `save` method of the `checkpoint` instance. Defaults to
        ``InMemory``.
    epoch_cb: EpochCallback, optional
        Callback called after each epoch with epoch, train_loss, test_loss,
        and current learning rate. Defaults to ``EpochPrinter``.
    train_cb: TrainCallback, optional
        Callback called after training finished with last epoch, epoch with the
        best loss, the best loss itself, whether `max_epochs` was exhausted,
        and with the training history in the form of a dictionary of lists
        with train loss, test loss and learning rate.
        Defaults to ``TrainPrinter``.

    Warnings
    --------
    optimizer
        Because the (partial) optimizer will simply be completed with
        ``model.parameters()``, parameter groups are not supported.
    scheduler
        Not all learning-rate schedulers are supported . Their ``step()``
        method is called only once after each epoch, and it is called without
        any arguments.

    See Also
    --------
    InMemory
    EpochCallback
    EpochPrinter
    TrainPrinter

    """

    def __init__(
            self,
            batch_size: int,
            max_epochs: int,
            loss: Module,
            optimizer: Curry[Optimizer],
            scheduler: Curry[LRScheduler] = Curry[NoSchedule](NoSchedule),
            warmup: int = 0,
            patience: int | None = None,
            max_n: int = None,
            checkpoint: Checkpoint = InMemory(),
            epoch_cb: EpochCallback = EpochPrinter(),
            train_cb: TrainCallback = TrainPrinter()
    ) -> None:
        self.batch_size = batch_size
        self.max_epochs = max_epochs
        self.loss = loss
        self.optimizer = optimizer
        self.scheduler = scheduler
        self.warmup = warmup
        self.patience = max_epochs if patience is None else patience
        self.max_n = max_n
        self.checkpoint = checkpoint
        self.epoch_cb = epoch_cb
        self.train_cb = train_cb
        super().__init__(
            batch_size,
            max_epochs,
            loss,
            optimizer,
            scheduler,
            warmup,
            patience,
            max_n,
            checkpoint,
            epoch_cb,
            train_cb
        )
        self.history = {'train_loss': [], 'test_loss': [], 'lr': []}

    def train(
            self,
            model: Module,
            train: TrainDataBase,
            test: TestDataBase | None = None
    ) -> Module:
        """Train a fresh model from scratch, starting from a clean slate.

        Parameters
        ----------
        model: Module
            PyTorch model to train. Must have a ``reset_parameters()`` method
            that can be called without any parameters to re-initialize all
            trainable model parameters and buffers.
        train: TrainDataBase
            Training data.
        test: TestDataBase, optional
            Hold-out data to compute test loss. If configured, early stopping
            will track the `test` loss if `test` is given and the `train` loss
            if it is not.

        Returns
        -------
        Module
            The trained Pytorch model.

        Raises
        ------
        TrainError
            If the `model` does not have a ``reset_parameters()`` method.

        Important
        ---------
        The `model` to train must always return a *tuple* of tensors, not just
        a single tensor. If it produces only one tensor, return it as a 1-tuple
        of tensors.

        Warnings
        --------
        Model training is (re-)started from scratch every time this method is
        called. The training history is erased, all internal model parameters
        are reset to a pristine initial state, and previously saved checkpoints
        are irrevocably deleted!

        """
        self.checkpoint.reset_parameters()
        self.history = {'train_loss': [], 'test_loss': [], 'lr': []}
        try:
            model.reset_parameters()
        except AttributeError as error:
            msg = 'Models must have a "reset_parameters()" method!"'
            raise TrainError(msg) from error
        return self.resume(model, train, test)

    def resume(
            self,
            model: Module,
            train: TrainDataBase,
            test: TestDataBase | None = None
    ) -> Module:
        """Resume model training from the best epoch checkpointed so far.

        Parameters
        ----------
        model: Module
            PyTorch model to train.
        train: TrainDataBase
            Training data.
        test: TestDataBase, optional
            Hold-out data to compute test loss. If configured, early stopping
            will track the `test` loss if `test` is given and the `train` loss
            if it is not.

        Returns
        -------
        Module
            The trained Pytorch model.

        Notes
        -----
        This is safe to use even if you have never trained your model before,
        and you are starting from scratch.

        Important
        ---------
        The `model` to train must always return a *tuple* of tensors, not just
        a single tensor. If it produces only one tensor, return it as a 1-tuple
        of tensors.

        """
        # How many data points to take for computing train (and test) loss.
        n = train.n if test is None else test.n
        self.max_n = n if self.max_n is None else self.max_n
        max_n = min(self.max_n, n)

        # Initialize training cycle.
        optimizer = self.optimizer(model.parameters())
        scheduler = self.scheduler(optimizer)
        epoch, best_loss = self.checkpoint.load(model, optimizer, scheduler)

        # Initialize counting and accumulation variables.
        best_epoch = epoch
        n_wait = 1
        max_epochs_reached = False

        # Loop over epochs.
        for epoch in range(epoch + 1, self.max_epochs + 1):
            # Train one epoch, looping over batches.
            model.train()
            for features, target in train(self.batch_size):
                optimizer.zero_grad()
                predictions = model(*features)
                loss = self.loss(*predictions, target)
                loss.backward()
                optimizer.step()

            # Evaluate model on train and, potentially, test data.
            model.eval()
            with pt.no_grad():
                features, target = train.sample(max_n)
                train_loss = self.loss(*model(*features), target).item()
                if test is None:
                    test_loss = float('nan')
                else:
                    features, target = test.sample(max_n)
                    test_loss = self.loss(*model(*features), target).item()

            # Append epoch metrics to training history.
            current_lr = scheduler.get_last_lr()[0]
            self.history['train_loss'].append(train_loss)
            self.history['test_loss'].append(test_loss)
            self.history['lr'].append(current_lr)

            # Call callback with epoch metrics and information.
            self.epoch_cb(
                epoch,
                train_loss,
                test_loss,
                current_lr,
                model,
                features,
                target
            )

            # Update the learning rate of the optimizer if warmup is exhausted.
            if epoch >= self.warmup:
                scheduler.step()

                # After warm-up, check if loss improved within our patience.
                track_loss = train_loss if test is None else test_loss
                if track_loss < best_loss:
                    best_loss = track_loss
                    best_epoch = epoch
                    n_wait = 1
                    self.checkpoint.save(
                        best_epoch,
                        best_loss,
                        model,
                        optimizer,
                        scheduler
                    )
                elif n_wait < self.patience:
                    n_wait += 1
                else:
                    self.checkpoint.load(model, optimizer, scheduler)
                    break

        # Did we exhaust the maximum number of epochs?
        else:
            max_epochs_reached = True

        # Call callbacks on finished training.
        self.epoch_cb.close()
        self.train_cb(
            epoch,
            best_epoch,
            best_loss,
            max_epochs_reached,
            self.history
        )

        # Finally, return the trained model.
        return model



import time
from typing import Any
from google.cloud.bigquery import Client, QueryJobConfig
from ...misc import ArgRepr
from .exceptions import GbqError


class GbqQuery(ArgRepr):
    """Run a SQL query that does not return anything on Google BigQuery.

    Parameters
    ----------
    project: str
        The name of the Google billing project.
    polling_interval: int, optional
        Job completion is going to be checked for every `polling_interval`
        seconds. Defaults to 5 (seconds).
    priority: str, optional
        Priority the query job should be run as. Can be either "BATCH" or
        "INTERACTIVE". Defaults to "BATCH". Use the `QueryPriority
        <https://cloud.google.com/python/docs/reference/bigquery/latest/
        google.cloud.bigquery.job.QueryPriority>`__ enum to avoid typos.
    **kwargs
        Additional keyword arguments are passed to the constructor of the
        Google BigQuery ``Client`` (see `documentation <https://cloud.google.
        com/python/docs/reference/bigquery/latest/google.cloud.bigquery.
        client.Client#parameters>`__ for options).

    Notes
    -----
    Typical use cases would be to create, move, alter, or delete tables.
    As such, the possibility to route the output of the query into a
    destination table is not foreseen.

    """

    def __init__(
            self,
            project: str,
            polling_interval: int = 5,
            priority: str = 'BATCH',
            **kwargs: Any
    ) -> None:
        self.project = project.strip(' /.')
        self.polling_interval = polling_interval
        self.priority = priority.strip().upper()
        self.kwargs = kwargs
        super().__init__(
            self.project,
            self.polling_interval,
            self.priority,
            **kwargs
        )

    def __call__(self, query: str, **kwargs: Any) -> tuple[()]:
        """Run a query that does not return anything on Google BigQuery.

        Parameters
        ----------
        query: str
            The SQL query to fire using the pre-configured client.
        **kwargs
            Additional keyword arguments are passed to the constructor of the
            Google BigQuery ``QueryJobConfig``. See `documentation
            <https://cloud.google.com/python/docs/reference/bigquery/latest/
            google.cloud.bigquery.job.QueryJobConfig>`__ for options.

        Returns
        -------
        tuple
            If the job finishes without errors, an empty tuple is returned.

        Raises
        ------
        GbqError
            If the ``QueryJob`` finishes and returns and error.

        """
        client = Client(self.project, **self.kwargs)
        config = QueryJobConfig(priority=self.priority, **kwargs)
        job = client.query(query, config)
        while job.running():
            time.sleep(self.polling_interval)
        if error := job.error_result:
            raise GbqError(f"\n{error['reason'].upper()}: {error['message']}")
        return ()



from collections.abc import Iterable, Callable
from ..misc import ArgRepr
from .exceptions import ReduceError


class Reduce[T, S](ArgRepr):
    """Equivalent to a partial of the ``functools.reduce`` function.

    Upon subclassing and/or instantiation, type annotation with the return type
    of `call` (which must be the same as the type of its first argument as well
    as the type of the accumulator `acc`) and the type ot its second argument
    (which must be the same as the type of the elements in the iterable acted
    upon) is recommended.

    Parameters
    ----------
    call: callable
        Callable accepting the current `acc` and the next element of iterable,
        returning the updated value of `acc`.
    acc: optional
        Initial value for the accumulator in the reduce operation. If not given
        calling reduce on an empty iterable will fail. Defaults to ``None``.

    """

    def __init__(
            self,
            call: Callable[[T, S], T],
            acc: T | None = None
    ) -> None:
        super().__init__(call, acc)
        self.call = call
        self.acc = acc

    def __call__(self, iterable: Iterable[S]) -> T:
        """Reduce an iterable by accumulating elements with the cached `call`.

        Parameters
        ----------
        iterable
            An iterable of elements to reduce.

        Returns
        -------
        object
            The reduced iterable.

        """
        iterator = iter(iterable)
        acc = next(iterator) if self.acc is None else self.acc
        for i, element in enumerate(iterator):
            try:
                acc = self.call(acc, element)
            except Exception as error:
                msg = '\n{} calling\n{}\non element #{}:\n{}\n{}'
                name = self._name(self.call)
                err_cls = error.__class__.__name__
                fmt = msg.format(err_cls, name, i + 1, element, error)
                raise ReduceError(fmt) from error
        return acc



"""Collection of convenience classes that do not fit any other category."""

from .repr import ArgRepr, IndentRepr
from .loggers import StdOutLogger, DEFAULT_FMT, PID_FMT

__all__ = [
    'ArgRepr',
    'IndentRepr',
    'StdOutLogger',
    'DEFAULT_FMT',
    'PID_FMT'
]



import json
from enum import StrEnum
from typing import Any, overload, Literal
from google.cloud.bigquery import Client, Dataset
from google.cloud.exceptions import NotFound
from google.api_core.retry import Retry

type CollationType = Literal['', 'und:ci']
type RoundingType = Literal['ROUND_HALF_AWAY_FROM_ZERO', 'ROUND_HALF_EVEN']
type BillingType = Literal['PHYSICAL', 'LOGICAL']


class Collation(StrEnum):
    """Specify string sorting in tables in a Google BigQuery dataset."""
    SENSITIVE = ''
    INSENSITIVE = 'und:ci'


class Rounding(StrEnum):
    """Specify rounding in tables in a Google BigQuery dataset."""
    HALF_AWAY = 'ROUND_HALF_AWAY_FROM_ZERO'
    HALF_EVEN = 'ROUND_HALF_EVEN'


class Billing(StrEnum):
    """Specify storage billing mode of tables in a Google BigQuery dataset."""
    PHYSICAL = 'PHYSICAL'
    LOGICAL = 'LOGICAL'


class GbqDataset:
    """Create a new dataset in a Google BigQuery project.

    Parameters
    ----------
    project: str
        The project to create the dataset in.
    dataset: str
        The identifier of the dataset to create. Only letters, numbers, and
        underscored are permitted.
    location: str
        The physical datacenter location to create the dataset in. See the
        Google Cloud Platform `documentation <https://cloud.google.com/
        bigquery/docs/locations>`__ for options.
    name: str, optional
        A human-readable name of the dataset. Defaults to the `dataset` it.
    description: str, optional
        A short description of the dataset. Defaults to ``None``.
    table_expire_days: int, optional
        Number of days after which tables are dropped. Defaults to ``None``,
        which results in tables never being dropped.
    partition_expire_days: int, optional
        Number of days after which partitions of partitioned tables are
        dropped. Defaults to ``None``, which results in partitions never
        being dropped.
    labels: dict, optional
        Any number of string-valued labels of the dataset. Defaults to none.
    access: list of dict, optional
        Fined-grained access rights to the dataset (see the Google Cloud
        Platform `documentation <https://cloud.google.com/bigquery/docs/
        reference/rest/v2/datasets#resource-dataset>`__ for details). If not
        given, defaults access rights are set by Google BigQuery.
    case_sensitive: bool, optional
        Whether dataset and table names should be case-sensitive or not.
        Defaults to `True`.
    collation: str, optional
        Default collation mode for string sorting in string columns of tables.
        Defaults to ``None``, which results in case-sensitive behavior. Use
        the ``Collation`` enum to specify explicitly.
    rounding: str, optional
        Default rounding mode. Defaults to ``None``, which lets Google BigQuery
        choose. Use the ``Rounding`` enum to specify explicitly.
    max_travel_time_hours: int, optional
        Define duration of Google Bigquery's "time travel" window in hours,
        i.e., for how long changes can be rolled back and tables can be
        queried "as of" some previous time. Values can be between 48 and 168
        hours (2 to 7 days). Defaults to 168.
    billing: str, optional
        Default billing mode for tables. Defaults to ``None``, which lets
        Google BigQuery choose. Use the ``Billing`` enum to specify explicitly.
    tags: dict, optional
        Associate globally defined tags with this dataset. Defaults to
        ``None``, which result in no tags to be associated.
    **kwargs
        Additional keyword arguments are passed to the constructor of the
        Google BigQuery ``Client`` (see `documentation <https://cloud.google.
        com/python/docs/reference/bigquery/latest/google.cloud.bigquery.
        client.Client#parameters>`__ for options).

    Notes
    -----
    Options for linked and external dataset sources, as well as for encryption
    configuration are deliberately omitted. You probably should not play with
    those without consulting your organization's data engineers.

    See Also
    --------
    Collation
    Rounding
    Billing

    """

    def __init__(
            self,
            project: str,
            dataset: str,
            location: str,
            name: str | None = None,
            description: str | None = None,
            table_expire_days: int | None = None,
            partition_expire_days: int | None = None,
            labels: dict[str, str] | None = None,
            access: list[dict[str, str]] | None = None,
            case_sensitive: bool = True,
            collation: CollationType | None = None,
            rounding: RoundingType | None = None,
            max_travel_time_hours: int = 168,
            billing: BillingType | None = None,
            tags: dict[str, str] | None = None,
            **kwargs: Any
    ) -> None:
        self.project = project.strip(' /.')
        self.dataset = dataset.strip(' /.')
        self.location = location.strip().lower()
        self.name = self.dataset if name is None else name.strip()
        self.description = description
        self.table_expire_days = table_expire_days
        self.partition_expire_days = partition_expire_days
        self.labels = {} if labels is None else labels
        self.access = access
        self.case_sensitive = case_sensitive
        self.collation = collation
        self.rounding = rounding
        self.max_travel_time_hours = max_travel_time_hours
        self.billing = billing
        self.tags = {} if tags is None else tags
        self.kwargs = kwargs

    @staticmethod
    @overload
    def to_ms(days: None) -> None:
        ...

    @staticmethod
    @overload
    def to_ms(days: int) -> int:
        ...

    @staticmethod
    def to_ms(days):
        """Convert integer days to millisecond string for the GCP API call."""
        return None if days is None else f'{int(days * 1000 * 60 * 60 * 24)}'

    @property
    def api_repr(self) -> dict[str, Any]:
        """Payload for the API call to the Google Cloud Platform."""
        return {
            'datasetReference': {
                'projectId': self.project,
                'datasetId': self.dataset
            },
            'friendlyName': self.name,
            'description': self.description,
            'defaultTableExpirationMs': self.to_ms(self.table_expire_days),
            'defaultPartitionExpirationMs': self.to_ms(
                self.partition_expire_days),
            'labels': self.labels,
            'access': self.access,
            'location': self.location,
            'isCaseInsensitive': not self.case_sensitive,
            'defaultCollation': self.collation,
            'defaultRoundingMode': self.rounding,
            'maxTimeTravelHours': f'{self.max_travel_time_hours}',
            'storageBillingModel': self.billing,
            'resourceTags': self.tags
        }

    def __repr__(self) -> str:
        return json.dumps(self.api_repr, indent=4)

    def __call__(
            self,
            exists_ok: bool = True,
            retry: Retry | None = None,
            timeout: float | tuple[float, float] | None = None,
    ) -> tuple[Dataset, bool]:
        """Create a Google BigQuery dataset in a Google Cloud Platform project.

        Parameters
        ----------
        exists_ok: bool, optional
            Whether to raise a ``Conflict`` exception if the targeted dataset
            already exists or not. Defaults to ``True``.
        retry: Retry, optional
            Retry policy for the request. Defaults to ``None``, which disables
            retries. See the Google Cloud Platform `guide
            <https://cloud.google.com/python/docs/reference/storage/1.39.0/
            retry_timeout#configuring-retries>`__ and `reference
            <https://googleapis.dev/python/google-api-core/latest/retry.html>`__
            for options.
        timeout: float, optional
            The number of seconds to wait for the HTTP response to the API call
            before using `retry` or a tuple with separate values for connection
            and request timeouts. Defaults to ``None``, meaning wait forever.

        Raises
        ------
        Conflict
            If `exists_ok` is set to ``False`` and the dataset already exists.

        Returns
        -------
        Dataset
            The existing or newly created dataset. If existing, then the
            dataset is returned unchanged, that is, none of the specified
            options are applied.
        bool
            ``True`` if the requested dataset is newly created and ``False``
            if an existing dataset is returned.

        """
        dataset = Dataset.from_api_repr(self.api_repr)
        client = Client(self.project, location=self.location, **self.kwargs)
        try:
            _ = client.get_dataset(dataset.reference, retry, timeout)
        except NotFound:
            exist = False
        else:
            exist = True
        return client.create_dataset(dataset, exists_ok, retry, timeout), exist



"""Thread- and process-parallel versions for elements of your workflow."""

from .threadmap import ThreadMap
from .threadfork import ThreadFork
from .processmap import ProcessMap
from .processfork import ProcessFork

__all__ = [
    'ThreadMap',
    'ThreadFork',
    'ProcessMap',
    'ProcessFork'
]

# ToDo: Add ThreadRoute and ProcessRoute
# ToDo: Add lazy subpackage



from typing import Any, Self
from collections.abc import KeysView, Callable, Iterator
from functools import reduce
from ast import literal_eval
import json
from json import JSONDecodeError
from pandas import Series
from .fields import Maybe
from .exceptions import (
    SchemaError,
    DefaultsError,
    ParseError,
    CastError,
    ValidationErrors
)

type Json = dict[str, Any]
type Raw = str | bytes | bytearray | Json | Series | None
type Schema = dict[str, type | Callable[[Any], Any]]
type Types = tuple[type, ...]
type Tried = tuple[Json, list[Exception]]


class SchemaMeta(type):
    """Metaclass parsing, validating, and setting data schema definitions.

    This class is not intended to be instantiated directly but only to be used
    as a metaclass! When used as such, it parses type-annotated class variables
    (with or without default values) into `schema` and `defaults` class
    variables, and checks that type annotations are callable, can be called
    on the default values, and that default values are not ``None`` unless
    wrapped with ``Maybe``. Additionally, it makes sure that schema fields
    cannot collide with existing class variables as well as instance attributes
    and methods.

    """

    __blacklist__ = {
        'as_json',
        'as_series',
        'as_dtype',
        'get',  # Do we even need this method?
        'keys',
        '_serialize'
    }

    __ignore_extra__ = False
    __raise_extra__ = True
    __respect_none__ = False

    def __new__(
            mcs,  # noqa: N804
            name: str,
            bases: Types,
            attrs: Json,
            **kwargs: Any
    ) -> 'SchemaMeta':
        # Create a new class object
        cls = super().__new__(mcs, name, bases, attrs)

        # Set behavior from past and present class keywords
        ancestral_variables = mcs.__ancestral(cls, '__dict__')
        cls.__ignore_extra__ = kwargs.pop(
            'ignore_extra',
            ancestral_variables.get('__ignore_extra__', mcs.__ignore_extra__)
        )
        cls.__raise_extra__ = kwargs.pop(
            'raise_extra',
            ancestral_variables.get('__raise_extra__', mcs.__raise_extra__)
        )
        cls.__respect_none__ = kwargs.pop(
            'respect_none',
            ancestral_variables.get('__respect_none__', mcs.__respect_none__)
        )

        # Consolidate values from "keys_from" keyword and the remaining kwargs
        kwargs = kwargs.pop('keys_from', {}) | kwargs

        # The schema is in the class __annotations__
        ancestral_schema = mcs.__ancestral(cls, '__annotations__')
        # Extract (string) keys from additional keyword arguments
        kwarg_schema = {key: str for key in kwargs}
        # Class body fields overwrite keyword fields overwrite inherited fields
        schema = ancestral_schema | kwarg_schema | cls.__annotations__
        # Validate the schema just assembled
        schema, schema_errors = mcs.__valid(name, schema, mcs.__blacklist__)

        # Ancestral defaults are in the class __defaults__
        ancestral_defaults = mcs.__ancestral(cls, '__defaults__')
        # Defaults for additional fields from keyword arguments are the keys
        kwarg_defaults = {k: k for k in kwargs if k not in cls.__annotations__}
        # Class variables overwrite keyword defaults overwrite ancestry
        updated_defaults = ancestral_defaults | kwarg_defaults | cls.__dict__
        # Only type-annotated class variables are relevant
        filtered_defaults = mcs.__filter(schema, updated_defaults)
        # Use __defaults__ because __dict__ cannot be set or updated directly
        defaults, default_errors = mcs.__tried(filtered_defaults, schema)

        # Raise accumulated errors, if any
        errors = schema_errors + default_errors
        if errors:
            raise ValidationErrors(name, errors)

        # Set hidden class variables
        cls.__blacklist__ = mcs.__blacklist__
        cls.__annotations__ = schema
        cls.__defaults__ = defaults

        return cls

    @staticmethod
    def __ancestral(descendant: 'SchemaMeta', attribute: str) -> Json:
        """Accumulate dictionary class variables down the inheritance tree."""
        # Get class ancestors starting with the oldest
        lineage = reversed(descendant.mro()[1:])
        # Accumulate inherited dictionary attributes, overwriting old with new
        return reduce(descendant.__merge(attribute), lineage, {})

    @staticmethod
    def __merge(attribute: str) -> Callable[[Json, Any], Json]:
        """Provide update function for dictionary class attributes."""

        def update(older: Json, newer: Any) -> Json:
            """Update dictionary attribute of parent with that of child."""
            return {**older, **getattr(newer, attribute, {})}

        return update

    @staticmethod
    def __filter(schema: Schema, defaults: Json) -> Json:
        """Filter down class __dict__ to keys present in the schema."""
        return {key: defaults[key] for key in defaults if key in schema}

    @staticmethod
    def __valid(
            name: str,
            schema: Schema,
            blacklist: set[str]
    ) -> tuple[Schema, list[Exception]]:
        """Validate that class-variable annotations are sane."""
        hidden = f'_{name}__'  # Pattern for double-underscore class variables
        errors = []
        for field, annotation in schema.items():
            if not callable(annotation):
                msg = f'Annotation of field "{field}" is not callable!'
                errors.append(SchemaError(msg))
            if field in blacklist:
                msg = f'Field "{field}" is on the blacklist {blacklist}!'
                errors.append(SchemaError(msg))
            if field.startswith(hidden):
                cleaned = field.removeprefix(hidden)
                msg = f'Field "__{cleaned}" starts with two underscores "__"!'
                errors.append(SchemaError(msg))
        return schema, errors

    @staticmethod
    def __tried(defaults: Json, schema: Schema) -> Tried:
        """Ensure that class-variable defaults are sane."""
        errors = []
        for item in defaults:
            # Check for None values and whether they are allowed
            default_is_none = defaults[item] is None
            type_is_not_maybe = not isinstance(schema[item], Maybe)
            if default_is_none and type_is_not_maybe:
                msg = (f'For the default value of field "{item} to be None,'
                       ' annotate it as Maybe(<YOUR_TYPE>) in the schema!')
                errors.append(DefaultsError(msg))
            # Check that schema annotations can be called on default values
            try:
                defaults[item] = schema[item](defaults[item])
            except (TypeError, ValueError):
                msg = (f'Default value for field "{item}" can'
                       ' not be cast to the desired type!')
                errors.append(DefaultsError(msg))
        return defaults, errors


class JsonObject(metaclass=SchemaMeta):
    """Flexible Dataclass-like data structure with enforced type schema.

    This class is not meant to ever be instantiated directly. Rather,
    inherit from it, and specify fields as type-annotated class variables,
    potentially giving also default values. Values for non-default fields
    must be provided on instantiation in the form of a JSON string, a
    dictionary-like object, or keyword arguments. The handling of additional
    fields can be specified via boolean class keywords `ignore_extra`
    (defaults to ``False``) and `raise_extra` (defaults to ``True``).

    By default, JSON fields with a ``None`` value are ignored and treated as
    not being present. To actually set fields to ``None`` (and, potentially,
    overwrite defaults), the class keyword `respect_none` needs to be set
    to ``True`` on subclass definition. Note, however, that type annotations
    must also tolerate ``None`` values, which is realized by wrapping existing
    types into ``Maybe`` instances.

    The resulting object behaves in many ways like a dictionary, allowing
    dictionary-style, but also object-style access to data fields. Attributes
    of nested instances can be accessed dictionary-style (i.e., with the
    square-bracket accessor) with a dot.separated key.

    Parameters
    ----------
    mapping: dict, str, bytes, or Series, optional
        Dictionary with string keys, JSON string/bytes, or pandas Series.
        Defaults to an empty dictionary.
    **kwargs
        Can be any value or, for nested structures, again a dictionary with
        string keys or a JSON string/bytes or a pandas Series. Keyword
        arguments will override values already present in the `mapping`.

    Raises
    ------
    ValidationErrors
        ExceptionGroup containing any number of the following exceptions.
    ParseError
        If the (keyword) arguments cannot be parsed into a dictionary with
        string keys and if non-default fields are neither given in the
        `mapping` nor in the keyword arguments.
    CastError
        If the dictionary values cannot be cast into the types specified in
        the schema.

    Warnings
    --------
    This class is rather heavy, so do not use it to, e.g., wrap JSON payloads
    in high-throughput low-latency web services!

    See Also
    --------
    fields.Maybe

    """

    def __init__(self, mapping: Raw | Self = None, **kwargs: Any) -> None:
        # Fully nest the parsed and purged dictionaries with dot.separated keys
        parsed = self.__nest(self.__purge(self.__parse(mapping)))
        kwargs = self.__nest(self.__purge(self.__parse(kwargs, 1)))
        defaults = self.__nest(self.__defaults__)
        # Merge the fully nested dictionaries
        merged = self.__merge(defaults, self.__merge(parsed, kwargs))
        # Type-cast the merged dictionary
        cast = self.__cast(merged)
        # Set all dictionary items as object attributes
        self.__dict__.update(cast)

    def __getitem__(self, key: str) -> Any:
        # Raise for blacklisted keys
        if key in self.__blacklist__:
            raise KeyError(key)
        # Try and split the (string) key by dots
        try:
            root, *children = key.split('.')
        except AttributeError as error:
            cls = type(key).__name__
            msg = f'Keys must be strings, not {cls} like {key}!'
            raise KeyError(msg) from error
        # The key could also refer to an attribute like a property or a method
        try:
            value = self.__dict__.get(root, getattr(self, root))
        # ... but we still raise a KeyError to meet expectations
        except AttributeError as error:
            raise KeyError(key) from error
        # If the key contains dots, recurse down into the value
        return reduce(lambda x, y: x[y], children, value)

    def __iter__(self) -> Iterator[str]:
        return self.__dict__.__iter__()

    def __str__(self) -> str:
        return json.dumps(self.__dict__, default=self._serialize)

    def __repr__(self) -> str:
        return json.dumps(self.__dict__, indent=4, default=self._serialize)

    def __eq__(self, other: Self) -> bool:
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        return NotImplemented

    def __ne__(self, other: Self) -> bool:
        if isinstance(other, self.__class__):
            return self.__dict__ != other.__dict__
        return NotImplemented

    def __contains__(self, key: str) -> bool:
        return key in self.__dict__

    def __len__(self) -> int:
        return len(self.__dict__)

    def __bool__(self) -> bool:
        return bool(self.__dict__)

    def __call__(self, mapping: Raw | Self = None, **kwargs: Any) -> Self:
        """Update one or more (nested) fields with `mapping` and kwargs.

        Parameters
        ----------
        mapping: dict, str, bytes, or Series, optional
            Dictionary with string keys, JSON string/bytes, or pandas Series.
            Defaults to an empty dictionary.
        **kwargs
            Can be any value or, for nested structures, again a dictionary with
            string keys or a JSON string/bytes or a pandas Series. Keyword
            arguments will override values already present in the `mapping`.

        Returns
        -------
        JsonObject
            A new instance of self with updated values.

        Raises
        ------
        ValidationErrors
            ExceptionGroup containing any number of the following exceptions.
        ParseError
            If the (keyword) arguments cannot be parsed into a dictionary with
            string keys.
        CastError
            If the dictionary values cannot be cast into the types specified in
            the schema.

        """
        # Fully nest the parsed and purged dictionaries with dot.separated keys
        parsed = self.__nest(self.__purge(self.__parse(mapping)))
        kwargs = self.__nest(self.__purge(self.__parse(kwargs, 1)))
        # Merge the call arguments
        merged = self.__merge(parsed, kwargs)
        # Left merge with self
        merged = self.__merge(self.__dict__, merged,True)
        # Instantiate a new, updated copy of self from the fully nested update
        return self.__class__(merged)

    @property
    def as_json(self) -> Json:
        """JSON-serializable dictionary representation."""
        return json.loads(str(self))

    @property
    def as_dtype(self) -> str:
        """Representation in a cell of a pandas data frame."""
        return self.__str__()

    @property
    def as_series(self) -> Series:
        """Representation as a pandas series."""
        data = {key: getattr(self[key], 'as_dtype', self[key]) for key in self}
        name = self.__class__.__name__
        return Series(data, name=name)

    # Do we even need this method?
    def get(self, item: str, default: Any = None) -> Any:
        """Get (nested) attribute by (dot.separated) name or default."""
        try:
            return self[item]
        except KeyError:
            return default

    def keys(self) -> KeysView[str]:
        """Attribute names as dictionary keys."""
        return self.__dict__.keys()

    @staticmethod
    def _serialize(obj: Any) -> Any:
        """Default JSON-encoding for attributes not trivially serializable."""
        return obj.as_json if hasattr(obj, 'as_json') else repr(obj)

    def __parse(self, obj: Raw | Self, level: int = 0) -> Json:
        """Recursively parse input into a (nested) dictionary."""
        # For the initial, root-level call, None means an empty dictionary
        mapping = {} if obj is None and level == 0 else obj
        # Try to parse the input as a JSON string ...
        try:
            parsed = json.loads(mapping)
        except (TypeError, JSONDecodeError):
            # ... or some other string representation of a python object
            try:
                parsed = literal_eval(mapping)
            # In case of failure, it might already be a dictionary
            except (TypeError, ValueError, SyntaxError):
                parsed = mapping
        # If it is, this should work.
        try:
            parsed = {**parsed}
        # If not ...
        except TypeError as error:
            # ... we're done with the recursion and simply return the input ...
            if level > 0:
                return obj
            # ... unless this was the initial, root-level call. Then we fail
            raise ParseError(f'Could not parse {obj} as JSON!') from error
        # Recurse further down into the value of the parsed dictionary
        return {key: self.__parse(parsed[key], level + 1) for key in parsed}

    def __purge(self, mapping: Json) -> Json:
        """Eliminate items with ``None`` value according to  `respect_none`."""
        filters = {True: lambda _: True, False: lambda xs: xs[1] is not None}
        return dict(filter(filters[self.__respect_none__], mapping.items()))

    @staticmethod
    def __stop_recursion_for(obj: Any) -> bool:
        """Criterion for stopping recursions dictionary nesting and merging.

        As we recursively traverse the tree of dictionary-like objects from
        root to leaves, we stop when we arrive at a leave that is no longer
        dictionary-like.

        """
        try:
            _ = [*obj.keys()]
        except (AttributeError, TypeError):
            return True
        return not hasattr(obj, '__getitem__')

    def __nest(self, mapping: Json | Self) -> Json:
        """Nest a dictionary with nesting implied by dot.separated keys."""
        # If the input is no longer dictionary-like, end the recursion
        if self.__stop_recursion_for(mapping):
            return mapping
        # If it is, initialize the return value ...
        result = {}
        # ... and iterate through the keys
        for key in mapping.keys():  # noqa: SIM118
            # Get the value to the current key
            value = mapping[key]
            # Depending on the type to key ...
            if isinstance(key, str):
                # ... split the root from the children
                root, *children = key.split('.')
            else:
                # ... or leave it as it is
                root, *children = key,
            # If the current key did have dots, ...
            if children:
                #  ... the value is elevated to a dict
                value = {'.'.join(children): value}
            # If the root key already exists in the results ...
            if root in result:
                # ... merge it with the new value
                result[root] = self.__merge(result[root], value)
            else:
                # ... or, if not, just set it to the new value
                result[root] = value
        # After nesting one level, recurse further down on the values
        return {key: self.__nest(value) for key, value in result.items()}

    def __merge(self, old: Json, new: Json, left: bool = False) -> Json:
        """Recursively deep-merge two dictionaries, outer or left."""
        if self.__stop_recursion_for(old) or self.__stop_recursion_for(new):
            return new
        # First the old values in order of appearance ...
        result = {key: old[key] for key in old if key not in new}
        # ... then intersection of old and new in order of appearance in old
        for key in [key for key in old if key in new]:
            result[key] = self.__merge(old[key], new[key], left)
        # If requested, add fields only present in new in order of appearance
        right = {} if left else {k: new[k] for k in new if k not in old}
        return result | right

    def __cast(self, mapping: Json) -> Json:
        """Cast all fields in the data structure to their specified type."""
        # Initialize accumulators
        cast = {}
        errors = []

        # Iterate over the fields in the schema
        for item, type_cast in self.__annotations__.items():
            try:
                value = mapping[item]
            except KeyError:
                msg = f'Missing non-default field "{item}"!'
                errors.append(ParseError(msg))
                continue
            try:
                cast[item] = type_cast(value)
            except (TypeError, ValueError):
                msg = f'Could not cast field "{item}" to the desired type!'
                errors.append(CastError(msg))
            except ValidationErrors as error_group:
                errors.append(error_group)
            if value is None and not isinstance(type_cast, Maybe):
                msg = (f'For the value of field "{item} to be None, annotate'
                       ' it as Maybe(<YOUR_TYPE>) in the schema!')
                errors.append(CastError(msg))

        # If we don't have to deal with extra fields, we're done
        if self.__ignore_extra__:
            if errors:
                raise ValidationErrors(self.__class__.__name__, errors)
            return cast

        # If not, first check if we even allow extra fields
        extra_fields = set(mapping) - set(self.__annotations__)
        if extra_fields and self.__raise_extra__:
            msg = f'Fields {extra_fields} are not in the schema!'
            errors.append(ParseError(msg))
            raise ValidationErrors(self.__class__.__name__, errors)

        # Even if extra fields are not ignored and are allowed, we need to ...
        for field in extra_fields:
            # ... check that their keys are strings, ...
            if not isinstance(field, str):
                msg = f'Extra field "{field}" does not have a string key!'
                errors.append(ParseError(msg))
                continue
            # ... check that their keys are not blacklisted, ...
            if field in self.__blacklist__:
                msg = (f'Extra field "{field}" is on the '
                       f'blacklist {self.__blacklist__}!')
                errors.append(ParseError(msg))
            # ... and check that their keys are valid python identifiers.
            if not all(part.isidentifier() for part in field.split('.')):
                msg = (f'Not all parts of the (potentially dot.separated) key'
                       f' of field "{field}" are valid python identifiers!')
                errors.append(ParseError(msg))

        # If we found anything fishy, raise all errors together
        if errors:
            raise ValidationErrors(self.__class__.__name__, errors)

        # Only now do we accept and merge extra fields.
        extras = {field: mapping[field] for field in extra_fields}
        return {**cast, **extras}



from abc import ABC, abstractmethod
from ..types import Batch, Batches

__all__ = [
    'TestDataBase',
    'TrainDataBase'
]


class TestDataBase(ABC):

    @property
    @abstractmethod
    def n(self) -> int:
        """The total number of data points."""
        ...

    @abstractmethod
    def sample(self, size: int) -> Batch:
        """Return a reproducible (sub-)sample of your data with the given size.

        Needed to consistently evaluate (and report) the train and/or test
        error after every epoch of training.

        Parameters
        ----------
        size: int
            Number of data points in (or `batch_size` of) the reproducible
            sample of your data. Every time `step` is called with the same
            `size`, the exact same output should be returned so that train and
            test errors can be used to track training convergence.

        Returns
        -------
        tuple
            A tuple of input tensors to call your model with.
        Tensor
            The matching target values. Must have the same dimensions and sizes
            as the output of your model.

        Important
        ---------
        Even if your model only takes a single tensor as input, this method
        must return a tuple to cover the general case of your model taking
        more than one tensor as input. If this is not the case, simple return
        a 1-tuple of tensors.

        """
        ...


class TrainDataBase(TestDataBase):

    @abstractmethod
    def __call__(self, batch_size: int) -> Batches:
        """Return an iterator over the mini-batches your model is trained on.

        Parameters
        ----------
        batch_size: int
            The (maximum) number of data points in one batch.

        Returns
        -------
        Iterator
            One element yielded by the iterator is a 2-tuple. The first element
            is again a tuple, containing the input tensor(s) to call your model
            with. The second element of the tuple is a single tensor with
            the matching target values. It must have the same dimensions and
            sizes as the output of your model.

        Important
        ---------
        Even if your model only takes a single tensor as input, the first
        element of one tuple yielded by the returned iterator must always be a
        tuple of tensors to cover the general case of your model taking more
        than one tensor as input. If this is not the case, simple make that a
        1-tuple of tensors!

        """
        ...



import sys
import logging
from functools import cached_property
from logging import Logger, Formatter, StreamHandler, Handler
from .repr import ArgRepr

DEFAULT_FMT = '{asctime:<23s} [{levelname:<8s}] {message} ({name})'
PID_FMT = '{asctime:<23s} [{levelname:<8s}] {message} ({name} | PID-{process})'


class StdOutLogger(ArgRepr):
    """Wrapped logger to stdout with at least one formatted StreamHandler.

    The instantiation of the actual, underlying Logger is delayed until it is
    first needed to facilitate usage in multiprocessing scenarios.

    Parameters
    ----------
    name: str
        Name of the Logger. Typically set to ``__name__``.
    level: int, optional
        Minimum logging level. Defaults to 10 (= DEBUG).
    fmt: str, optional
        Format string for the log messages in ``str.format()`` format.

    Warnings
    --------
    To avoid creating and adding new Handler every time the same Logger is
    requested, one of its existing StreamHandlers will be modified if there
    are any. A new one will be created and added only if there aren't.
    By consequence, requesting the same Logger multiple times with a different
    `level` and/or a different `fmt` might change that Logger globally.

    """
    def __init__(
            self,
            name: str,
            level: int = logging.DEBUG,
            fmt: str = DEFAULT_FMT
    ) -> None:
        super().__init__(name, level, fmt)
        self.name = name
        self.level = level
        self.fmt = fmt

    def log(self, level: int, message: str) -> tuple[()]:
        """Log a message at the given level.

        Parameters
        ----------
        level: int
            The level to log at.
        message: str
            The message to log.

        Returns
        -------
        tuple
            An empty tuple.

        """
        self.logger.log(level, message)
        return ()

    def debug(self, message: str) -> tuple[()]:
        """Log a DEBUG message.

        Parameters
        ----------
        message: str
            The message to log.

        Returns
        -------
        tuple
            An empty tuple.

        """
        return self.log(logging.DEBUG, message)

    def info(self, message: str) -> tuple[()]:
        """Log an INFO message.

        Parameters
        ----------
        message: str
            The message to log.

        Returns
        -------
        tuple
            An empty tuple.

        """
        return self.log(logging.INFO, message)

    def warning(self, message: str) -> tuple[()]:
        """Log a WARNING message.

        Parameters
        ----------
        message: str
            The message to log.

        Returns
        -------
        tuple
            An empty tuple.

        """
        return self.log(logging.WARNING, message)

    def error(self, message: str) -> tuple[()]:
        """Log an ERROR message.

        Parameters
        ----------
        message: str
            The message to log.

        Returns
        -------
        tuple
            An empty tuple.

        """
        return self.log(logging.ERROR, message)

    def critical(self, message: str) -> tuple[()]:
        """Log a CRITICAL message.

        Parameters
        ----------
        message: str
            The message to log.

        Returns
        -------
        tuple
            An empty tuple.

        """
        return self.log(logging.CRITICAL, message)

    @cached_property
    def logger(self) -> Logger:
        """The specified Logger with one StreamHandler configured to specs."""
        # Get logger with the given name
        logger = logging.getLogger(self.name)
        # Adjust the logger level so that messages from the handler get through
        logger.setLevel(min(max(self.level, logging.DEBUG), logging.CRITICAL))
        # Get StreamHandlers to stdout
        handlers = self.__filtered(logger.handlers)
        # Get the first matching handler if there is one or make a new one
        handler = handlers[0] if handlers else StreamHandler(sys.stdout)
        # Configure that handler according to specs
        configured = self.__configure(handler)
        # If the now configured handler was newly created, add it to the logger
        if not handlers:
            logger.addHandler(configured)
        return logger

    def __filtered(self, handlers: list[Handler]) -> tuple[StreamHandler, ...]:
        """Filter the Logger's Handlers according to the filter criterion."""
        return tuple(filter(self.__is_stdout_streamhandler, handlers))

    @staticmethod
    def __is_stdout_streamhandler(handler: Handler) -> bool:
        """Filter criterion for the Logger's Handlers."""
        if isinstance(handler, StreamHandler):
            return handler.stream is sys.stdout
        return False

    def __configure(self, handler: StreamHandler) -> StreamHandler:
        """Configure the selected StreamHandler to stdout."""
        # Set the level of the handler
        handler.setLevel(min(max(self.level, logging.DEBUG), logging.CRITICAL))
        # Set the configured formatter
        handler.setFormatter(self.__formatter)
        return handler

    @property
    def __formatter(self) -> Formatter:
        """LogRecord Formatter for the specified format."""
        formatter = Formatter(fmt=self.fmt, style='{')
        formatter.default_msec_format = '%s.%03d'
        return formatter



"""Tools to interact with elements of the Google Cloud Project (GCP).

Specifically, data scientists tend to interact mostly with Google's BigQuery
(BQ) data-warehouse solution and the Google Cloud Storage (GCS).

"""

from .dataset import Collation, Rounding, Billing, GbqDataset
from .bucket import Storage, GcsBucket
from .query import GbqQuery
from .query2gcs import GbqQuery2GcsParquet
from .gcs2local import GcsDir2LocalDir
from .gcs2df import GcsParquet2DataFrame
from .query2df import GbqQuery2DataFrame
from .df2gbq import IfExists, DataFrame2Gbq

__all__ = [
    'Collation',
    'Rounding',
    'Billing',
    'GbqDataset',
    'Storage',
    'GcsBucket',
    'GbqQuery',
    'GbqQuery2GcsParquet',
    'GcsDir2LocalDir',
    'GcsParquet2DataFrame',
    'GbqQuery2DataFrame',
    'IfExists',
    'DataFrame2Gbq'
]



"""Combine your embedded features into vectors of size model-dimension.

Two ways of doing this are provided here. One is to form a (weighted) sum
of the feature embeddings. When these weights are learnable themselves, they
can be interpreted as feature importance. The other is to concatenate the
embedding vectors of all features into a single, wide vector and to then
project it down again into a space with the same dimensions of the embedding
space.

"""

from .activated import ActivatedConcatMixer
from .gated import GatedConcatMixer
from .gated_residual import GatedResidualConcatMixer


__all__ = [
    'ActivatedConcatMixer',
    'GatedConcatMixer',
    'GatedResidualConcatMixer'
]



from typing import Any, Self, ParamSpec
from collections.abc import Iterator, Callable, Iterable
from functools import singledispatchmethod
from concurrent.futures import ProcessPoolExecutor
from ...misc import IndentRepr
from ..exceptions import ForkError

P = ParamSpec('P')
type Call = type | Callable[P, Any]


class ProcessFork[**P, T](IndentRepr):
    """Call multiple callables with the same argument(s) in parallel processes.

    Generic type annotation of instances is recommended. Provide a list of
    one or more input types that all callables take, followed by a ``tuple``
    specifying the concatenation of the return types of all callables, ignoring
    empty tuples. If only a single object remains, the type of that object
    should be annotated.


    Parameters
    ----------
    call: callable or iterable of callables, optional
        One callable or an iterator of callables to all call with the same
        arguments. Defaults to an empty tuple.
    *calls: callable
        Additional callables to call with the same argument(s).
    max_workers: int, optional
        Maximum number of worker processes used in the pool to call
        `calls` asynchronously. Will be forwarded to the constructor of
        ``ProcessPoolExecutor``. Defaults to 4.
    initializer: callable, optional
        Called at the start of each worker process. Will be forwarded to the
        constructor of ``ProcessPoolExecutor``. Defaults to ``None``.
    initargs: tuple, optional
        Arguments passed to the initializer. Will be forwarded to the
        constructor of ``ProcessPoolExecutor``. Defaults to an empty tuple.
    max_tasks_per_child: int, optional
        Maximum number of iterable items to transform in each worker process
        before they are being restarted. Defaults to ``None``, indicating no
        restart(s) at all.
    timeout: int or float, optional
        Maximum time (in seconds) to wait for results to be available. Defaults
        to ``None``, which means there is no limit for the time to wait. Will
        be forwarded to the ``result`` method of ``Future``.

    See Also
    --------
    concurrent.futures.ProcessPoolExecutor

    """

    def __init__(
            self,
            call: Call | Iterable[Call] = (),
            *calls: Call,
            max_workers: int | None = 4,
            initializer: Callable[..., Any] | None = None,
            initargs: tuple[Any, ...] = (),
            max_tasks_per_child: int | None = None,
            timeout: float | None = None
    ) -> None:
        self.calls = self.__valid(call) + self.__valid(calls)
        self.max_workers = max_workers
        self.initializer = initializer
        self.initargs = initargs
        self.max_tasks_per_child = max_tasks_per_child
        self.timeout = timeout
        super().__init__(
            self.calls,
            max_workers,
            initializer,
            initargs,
            max_tasks_per_child,
            timeout
        )

    def __iter__(self) -> Iterator[Call]:
        # We could also iterate over instances of self ...
        return self.calls.__iter__()

    def __len__(self) -> int:
        return self.calls.__len__()

    def __bool__(self) -> bool:
        return bool(self.calls)

    def __contains__(self, call: Call) -> bool:
        return call in self.calls

    def __reversed__(self) -> Self:
        return self.__class__(*reversed(self.calls))

    @singledispatchmethod
    def __getitem__(self, index: int) -> Call:
        # We could also return instances of self ...
        return self.calls[index]

    @__getitem__.register
    def _(self, index: slice) -> Self:
        return self.__class__(*self.calls[index])

    def __eq__(self, other: Self) -> bool:
        if isinstance(other, self.__class__):
            return self.calls == other.calls
        return NotImplemented

    def __ne__(self, other: Self) -> bool:
        if isinstance(other, self.__class__):
            return self.calls != other.calls
        return NotImplemented

    def __add__(self, other: Call | Iterable[Call] | Self) -> Self:
        if isinstance(other, self.__class__):
            return self.__class__(self.calls, *other.calls)
        try:
            return self.__class__(self.calls, *self.__valid(other))
        except ForkError:
            return NotImplemented

    def __radd__(self, other: Call | Iterable[Call] | Self) -> Self:
        if isinstance(other, self.__class__):
            return self.__class__(other.calls, *self.calls)
        try:
            return self.__class__(self.__valid(other), *self.calls)
        except ForkError:
            return NotImplemented

    def __call__(self, *args: P.args) -> T:
        """Concurrently call all specified `calls` with the same argument(s).

        Parameters
        ----------
        *args
            Arguments to call all `calls` with.

        Returns
        -------
        tuple or object
            Concatenation of all return values of all `calls` in order. If only
            one of the `calls` returns something other than an empty tuple,
            that object is returned.

        Raises
        ------
        ForkError
            When one of the `calls` raises an exception.

        """
        with ProcessPoolExecutor(
                self.max_workers,
                None,
                self.initializer,
                self.initargs,
                max_tasks_per_child=self.max_tasks_per_child
        ) as pool:
            futures = [pool.submit(call, *args) for call in self.calls]
            results = []
            for i, future in enumerate(futures):
                try:
                    result = future.result(self.timeout)
                except Exception as error:
                    msg = '\n{} executing\n{}\nin fork {} of\n{}\n{}'
                    err_cls = error.__class__.__name__
                    name = self._name(self.calls[i])
                    fmt = msg.format(err_cls, name, i, self, error)
                    raise ForkError(fmt) from error
                else:
                    if isinstance(result, tuple):
                        results.extend(result)
                    else:
                        results.append(result)
        return results[0] if len(results) == 1 else tuple(results)

    @staticmethod
    def __valid(calls: Call | Iterable[Call]) -> tuple[Call, ...]:
        """Ensure that the argument is indeed an iterable of callables."""
        if callable(calls):
            return calls,
        iterable = True
        all_callable = False
        try:
            all_callable = all(callable(call) for call in calls)
        except TypeError:
            iterable = False
        if iterable and all_callable:
            return tuple(calls)
        raise ForkError('All branches in the process-fork must be callable!')



from collections.abc import Iterable, Callable
from ..misc import ArgRepr
from .exceptions import MapError


class Map[**P, S, T](ArgRepr):
    """Equivalent to a partial of the python builtin ``map`` function.

    Upon subclassing and/or instantiation, type annotation with a list of the
    argument type(s) of `transform`, the return type of `call`, and the return
    type of `wrapper` is recommended.

    Parameters
    ----------
    transform: callable
        Transforms element(s) of the input iterable(s).
    wrapper: type or callable, optional
        If not given, an attempt will be made to return the type of the first
        iterable the callable instance is being called with (by calling its
        class with a list of the mapped elements). If explicitly given,
        `wrapper` will be called with a list mapped elements. Consequently,
        the return type will be the (return) type of `wrapper`.

    Notes
    -----
    In contrast to python's builtin lazy ``map`` function, which returns a
    generator object, the mapped iterable is fully manifested first and only
    then wrapped.

    """

    def __init__(
            self,
            transform: type[S] | Callable[P, S],
            wrapper: type[T] | Callable[[list[S]], T] | None = None
    ) -> None:
        super().__init__(transform, wrapper)
        self.transform = transform
        self.wrapper = wrapper

    def __call__(self, iterable: Iterable, *iterables: Iterable) -> T:
        """Transform the element(s) of the given iterable(s).

        Parameters
        ----------
        iterable: Iterable
            An iterable of elements to transform.
        *iterables: Iterable
            If given, the cached `transform` is called with the corresponding
            elements of `ìterable` and all `iterables` as arguments.

        Returns
        -------
        Sequence
            Same type as `iterable` if `wrapper` was not specified on
            instantiation or the (return) type of `wrapper`. Note that, as with
            python's builtin ``map`` function, the length of the output
            sequence is limited by the shortest of the input iterables.

        Raises
        ------
        MapError
            If calling the cached `transform` on any element(s) of the given
            iterable(s) raises an exception or if wrapping the results leads
            to an exception.

        """
        mapped = []
        for i, elements in enumerate(zip(*(iterable, *iterables))):
            try:
                mapped.append(self.transform(*elements))
            except Exception as error:
                msg = '\n{} calling\n{}\non element #{}:\n{}\n{}'
                name = self._name(self.transform)
                err_cls = error.__class__.__name__
                display = elements[0] if len(elements) == 1 else elements
                fmt = msg.format(err_cls, name, i, display, error)
                raise MapError(fmt) from error
        wrap = iterable.__class__ if self.wrapper is None else self.wrapper
        try:
            wrapped = wrap(mapped)
        except Exception as error:
            msg = '\n{} calling wrapper\n{}\non map results:\n{}'
            name = self._name(wrap)
            err_cls = error.__class__.__name__
            raise MapError(msg.format(err_cls, name, error)) from error
        return wrapped



from typing import TypedDict
from abc import ABC, abstractmethod
from collections.abc import Callable
from ...misc import ArgRepr
from ..types import Tensor, Tensors, Module

__all__ = [
    'EpochCallback',
    'EpochPrinter',
    'TrainCallback',
    'History',
    'TrainPrinter'
]

type TrainCallback = Callable[[int, int, float, bool, History], None]


class History(TypedDict):
    """Summary of metrics passed to the callback when training is finished."""
    train_loss: list[float]  #: List of losses evaluated on train data.
    test_loss: list[float | None]  #: List of losses evaluated on test data.
    lr: list[float]  #: List of learning rates.


class EpochCallback(ABC):
    """Base class to inherit from when implementing custom epoch callbacks."""

    @abstractmethod
    def __call__(
            self,
            epoch: int,
            train_loss: int,
            test_loss: int,
            learning_rate: float,
            model: Module,
            features: Tensors,
            target: Tensor
    ) -> None:
        """Called after each epoch to print, log, or otherwise analyse metrics.

        Parameters
        ----------
        epoch: int
            The current epoch in the training loop.
        train_loss: float
            The loss computed on a sample of the train data after the
            current `epoch`.
        test_loss: float
            The loss computed on a sample of the test data after the
            current `epoch`. Always ``nan`` if no test data is used.
        learning_rate: float
            Learning rate of the optimizer in the current `epoch`.
        model: Module
            A reference to the model being trained.
        features: tuple
            A (potentially 1-)tuple of tensors with features to call the
            `model` with. Sampled from the test data if present and from the
            train data otherwise.
        target: Tensor
            Target matching the `features`.

        """
        ...

    @abstractmethod
    def close(self) -> None:
        """Called at the end of training to be compatible with TensorBoard."""
        ...


class EpochPrinter(ArgRepr, EpochCallback):
    """Epoch callback assembling an informative message on training progress.

    Parameters
    ----------
    printer: callable, optional
        Will be called with the assembled message. Default to the python
        builtin ``print`` function, but could also be a logging command.

    """

    def __init__(self, printer: Callable[[str], None] = print) -> None:
        super().__init__(printer)
        self.printer = printer

    def __call__(
            self,
            epoch: int,
            train_loss: float,
            test_loss: float,
            learning_rate: float,
            model: Module,
            features: Tensors,
            target: Tensor
    ) -> None:
        """Assemble training-progress message and call the printer with it."""
        msg = (f'Epoch: {epoch:>4} | learning rate: {learning_rate:7.5f} | '
               f'train loss: {train_loss:7.5f} | test loss: {test_loss:7.5f}')
        return self.printer(msg)

    def close(self) -> None:
        """Does nothing because there is nothing to close."""


class TrainPrinter(ArgRepr):
    """Train callback assembling an informative message when training ends.

    Parameters
    ----------
    printer: callable, optional
        Will be called with the assembled message. Default to the python
        builtin ``print`` function, but could also be a logging command.

    """
    def __init__(self, printer: Callable[[str], None] = print) -> None:
        super().__init__(printer)
        self.printer = printer

    def __call__(
            self,
            epoch: int,
            best_epoch: int,
            best_loss: float,
            max_epochs_reached: bool,
            history: History
    ) -> None:
        """Assemble and print a summary of model training.

        Parameters
        ----------
        epoch: int
            The last epoch in the training loop.
        best_epoch: int
            The epoch with the lowest loss encountered.
        best_loss: float
            The lowest loss encountered.
        max_epochs_reached: bool
            Whether the maximum number of epochs was exhausted or not.
        history: History
            Dictionary with lists of train losses, test losses, and learning
            rates.

        """
        if max_epochs_reached:
            self.printer(f'Maximum number of {epoch} epochs exhausted!')
        else:
            msg = (f'Stopping after {epoch} epochs because, even after '
                   f'{epoch - best_epoch} epochs, the loss did not drop '
                   f'below the lowest value of {best_loss:7.5f} seen in epoch '
                   f'{best_epoch}. Recovering checkpoint from that epoch.')
            self.printer(msg)



import sys
import logging
from typing import ParamSpec
from collections.abc import Callable
from functools import cached_property
from logging import Logger, Formatter, StreamHandler, Handler
from ...misc import ArgRepr

P = ParamSpec('P')
type Message = str | Callable[P, str]

DEFAULT_FMT = '{asctime:<23s} [{levelname:<8s}] {message} ({name})'
PID_FMT = '{asctime:<23s} [{levelname:<8s}] {message} ({name} | PID-{process})'


class PassThroughStdOut(ArgRepr):
    """Pass-through Logger to stdout with at least one formatted StreamHandler.

    The instantiation of the actual, underlying Logger is delayed until it is
    first needed to facilitate usage in multiprocessing scenarios.

    Parameters
    ----------
    name: str
        Name of the Logger. Typically set to ``__name__``.
    level: int, optional
        Minimum logging level. Defaults to 10 (= DEBUG).
    fmt: str, optional
        Format string for the log messages in ``str.format()`` format.

    Warnings
    --------
    To avoid creating and adding new Handler every time the same Logger is
    requested, one of its existing StreamHandlers will be modified if there
    are any. A new one will be created and added only if there aren't.
    By consequence, requesting the same Logger multiple times with a different
    `level` and/or a different `fmt` will change that Logger globally.

    """
    def __init__(
            self,
            name: str,
            level: int = logging.DEBUG,
            fmt: str = DEFAULT_FMT
    ) -> None:
        super().__init__(name, level, fmt)
        self.name = name
        self.level = level
        self.fmt = fmt

    class Log:
        """Return type of the logging methods.

        Notes
        -----
        This class is not intended to ever be instantiated directly.

        See Also
        --------
        PassThroughStdOut.debug
        PassThroughStdOut.info
        PassThroughStdOut.warning
        PassThroughStdOut.error
        PassThroughStdOut.critical

        """

        def __init__(
                self,
                parent: 'PassThroughStdOut',
                level: int,
                msg: Message
        ) -> None:
            self.parent = parent
            self.level = level
            self.msg = msg

        def __call__(self, *args: P.args) -> P.args:
            """Log the cached message and pass through the input argument(s).

            Parameters
            ----------
            *args
                If `msg` is a string, these arguments are inconsequential,
                and it will simply be logged. If, however, the `msg` is
                callable, it will be called with these arguments and the
                return value will be logged instead.

            Returns
            -------
            object or tuple
                Single argument if called with only one, otherwise all of them.

            """
            msg = self.msg(*args) if callable(self.msg) else self.msg
            self.parent.logger.log(self.level, msg)
            return args[0] if len(args) == 1 else args

    def debug(self, message: Message) -> Log:
        """Log a DEBUG message, optionally depending on call arguments.

        Parameters
        ----------
        message: str or callable
            Message to log when the returned object is called. If it is
            callable, then it will be called with whatever argument(s) the
            returned object is called with and the result will be logged.

        Returns
        -------
        Log
            A callable object that logs the `message` when called.

        """
        return self.Log(self, logging.DEBUG, message)

    def info(self, message: Message):
        """Log an INFO message, optionally depending on call arguments.

        Parameters
        ----------
        message: str or callable
            Message to log when the returned object is called. If it is
            callable, then it will be called with whatever argument(s) the
            returned object is called with and the result will be logged.

        Returns
        -------
        Log
            A callable object that logs the `message` when called.

        """
        return self.Log(self, logging.INFO, message)

    def warning(self, message: Message):
        """Log a WARNING message, optionally depending on call arguments.

        Parameters
        ----------
        message: str or callable
            Message to log when the returned object is called. If it is
            callable, then it will be called with whatever argument(s) the
            returned object is called with and the result will be logged.

        Returns
        -------
        Log
            A callable object that logs the `message` when called.

        """
        return self.Log(self, logging.WARNING, message)

    def error(self, message: Message):
        """Log an ERROR message, optionally depending on call arguments.

        Parameters
        ----------
        message: str or callable
            Message to log when the returned object is called. If it is
            callable, then it will be called with whatever argument(s) the
            returned object is called with and the result will be logged.

        Returns
        -------
        Log
            A callable object that logs the `message` when called.

        """
        return self.Log(self, logging.ERROR, message)

    def critical(self, message: Message):
        """Log a CRITICAL message, optionally depending on call arguments.

        Parameters
        ----------
        message: str or callable
            Message to log when the returned object is called. If it is
            callable, then it will be called with whatever argument(s) the
            returned object is called with and the result will be logged.

        Returns
        -------
        Log
            A callable object that logs the `message` when called.

        """
        return self.Log(self, logging.CRITICAL, message)

    @cached_property
    def logger(self) -> Logger:
        """The specified Logger with one StreamHandler configured to specs."""
        # Get logger with the given name
        logger = logging.getLogger(self.name)
        # Adjust the logger level so that messages from the handler get through
        logger.setLevel(min(max(self.level, logging.DEBUG), logging.CRITICAL))
        # Get StreamHandlers to stdout
        handlers = self.__filtered(logger.handlers)
        # Get the first matching handler if there is one or make a new one
        handler = handlers[0] if handlers else StreamHandler(sys.stdout)
        # Configure that handler according to specs
        configured = self.__configure(handler)
        # If the now configured handler was newly created, add it to the logger
        if not handlers:
            logger.addHandler(configured)
        return logger

    def __filtered(self, handlers: list[Handler]) -> tuple[StreamHandler, ...]:
        """Filter the Logger's Handlers according to the filter criterion."""
        return tuple(filter(self.__is_stdout_streamhandler, handlers))

    @staticmethod
    def __is_stdout_streamhandler(handler: Handler) -> bool:
        """Filter criterion for the Logger's Handlers."""
        if isinstance(handler, StreamHandler):
            return handler.stream is sys.stdout
        return False

    def __configure(self, handler: StreamHandler) -> StreamHandler:
        """Configure the selected StreamHandler to stdout."""
        # Set the level of the handler
        handler.setLevel(min(max(self.level, logging.DEBUG), logging.CRITICAL))
        # Set the configured formatter
        handler.setFormatter(self.__formatter)
        return handler

    @property
    def __formatter(self) -> Formatter:
        """LogRecord Formatter for the specified format."""
        formatter = Formatter(fmt=self.fmt, style='{')
        formatter.default_msec_format = '%s.%03d'
        return formatter



"""Custom loss functions not (yet) shipped with PyTorch  out of the box.

Especially when networks should make probabilistic predictions, that is, when
they are set up to predict the parameters of an analytical probability mass or
density function instead of just an expectation value, special loss functions
are required that implement the negative log-likelihood of the matching
probability distribution.

"""

from enum import StrEnum
import math
import torch as pt
import torch.nn as ptn
import torch.special as pts
from .types import Tensor, Module
from .exceptions import LossError
from .misc import identity

__all__ = [
    'Reduction',
    'RMSELoss',
    'TweedieLoss',
    'BetaBernoulliLoss',
    'GammaLoss',
    'StudentLoss',
    'NegativeBinomialLoss'
]


class Reduction(StrEnum):
    """Specify the aggregation level when evaluating loss functions."""
    MEAN = 'mean'
    SUM = 'sum'
    NONE = 'none'


class _BaseLoss(Module):
    """Base class for custom loss functions.

    This is class is not meant to ever be instantiated directly. Rather,
    it is intended to be inherited from when implementing a custom loss
    function. Its purpose is mimicking the call signature of PyTorch builtin
    loss functions so that only the `forward` method needs to be implemented
    in subclasses.

    Parameters
    ----------
    reduction: string, optional
        One of "mean", "sum" or "none". Defaults to "mean". Whether and, if so,
        how to aggregate the tensor resulting from evaluating the point-wise
        loss function on the input. Use the ``Reduction`` enum to avoid typos.
    eps: float, optional
        Many loss functions require input and/or target tensors to be bound
        by some lower and/or upper value. It is the user's responsibility to
        ensure that they are. However, evaluating a loss function just at the
        interval boundary of its support might lead to numerical inaccuracies.
        To avoid these, it is often advisable to shift such values away from
        boundaries by a small value `eps`. Defaults to 1e-6.

    See Also
    --------
    Reduction

    """

    __reductions = {
        'mean': pt.mean,
        'sum': pt.sum,
        'none': identity
    }

    def __init__(self, reduction: str = 'mean', eps: float = 1e-6) -> None:
        super().__init__()
        self.reduction = reduction.strip().lower()
        self.register_buffer('eps', pt.tensor(eps), False)
        try:
            self._reduce = self.__reductions[self.reduction]
        except KeyError:
            msg = '"reduction" must be one of "mean", "sum", or "none"!'
            raise LossError(msg)

    def forward(self, *tensors: Tensor) -> Tensor:
        """Forward pass for custom losses. Implement in subclasses!"""
        raise NotImplementedError('Subclasses must implement a forward pass!')


class RMSELoss(Module):
    """Root mean squared error loss.

    PyTorch only comes with a mean squared error loss. Since it is often more
    intuitive to compare error and target value when they are on the same
    scale, the square root of the MSE is naively implemented here.

    Parameters
    ----------
    reduction: string, optional
        One of "mean", "sum" or "none". Defaults to "mean". Whether and, if so,
        how to aggregate the tensor resulting from evaluating the point-wise
        loss function on the input. Use the ``Reduction`` enum to avoid typos.
    eps: float, optional
        Many loss functions require input and/or target tensors to be bound
        by some lower and/or upper value. It is the user's responsibility to
        ensure that they are. However, evaluating a loss function just at the
        interval boundary of its support might lead to numerical inaccuracies.
        To avoid these, it is often advisable to shift such values away from
        boundaries by a small value `eps`. Defaults to 1e-8.

    See Also
    --------
    Reduction

    """

    def __init__(self, reduction: str = 'mean', eps: float = 1e-8) -> None:
        super().__init__()
        self.reduction = reduction
        self.register_buffer('eps', pt.tensor(eps), False)
        self.mse = ptn.MSELoss(reduction=reduction)

    def forward(self, y_hat: Tensor, y_true: Tensor) -> Tensor:
        """Compute RMSE loss between predicted and observed values.

        For numerical stability, MSE values smaller than `eps` will be
        clamped to `eps` before taking the square root.

        Parameters
        ----------
        y_hat: Tensor
            Predicted expectation values.
        y_true: Tensor
            Actually observed values. Should have the same dimensionality
            as `y_hat`, but this will not be checked for, let alone enforced.

        Returns
        -------
        Tensor
            The root mean square error.

        """
        return self.mse(y_hat, y_true).clamp(self.eps).sqrt()


class TweedieLoss(_BaseLoss):
    """Tweedie loss for zero-inflated, right-skewed, long-tailed target.

    Implements the `deviance <https://en.wikipedia.org/wiki/Deviance
    _(statistics)>`__ of the `Tweedie distribution <https://en.wikipedia.org/
    wiki/Tweedie_distribution>`__ for power parameters between 1 and 2, where
    it can be seen as a compound Poisson-Gamma distribution.

    Parameters
    ----------
    reduction: string, optional
        One of "mean", "sum" or "none". Defaults to "mean". Whether and, if so,
        how to aggregate the tensor resulting from evaluating the point-wise
        loss function on the input. Use the ``Reduction`` enum to avoid typos.
    eps: float, optional
        Many loss functions require input and/or target tensors to be bound
        by some lower and/or upper value. It is the user's responsibility to
        ensure that they are. However, evaluating a loss function just at the
        interval boundary of its support might lead to numerical inaccuracies.
        To avoid these, it is often advisable to shift such values away from
        boundaries by a small value `eps`. Defaults to 1e-6.

    See Also
    --------
    Reduction

    """

    def __init__(self, reduction: str = 'mean', eps: float = 1e-6) -> None:
        super().__init__(reduction, eps)

    def forward(self, mu: Tensor, p: Tensor, y_true: Tensor) -> Tensor:
        """Forward pass for the Tweedie loss function.

        Parameters
        ----------
        mu: Tensor
            Predicted expectation values. Should be greater than or equal
            to 0, but this will not be checked for, let alone enforced.
            However, `eps` will be added to ensure numerical stability.
        p: Tensor
            Predicted values for the power parameter. Should have values
            strictly between 1 and 2. All values beyond that range will be
            clamped to lie within this range (by at least `eps`).
        y_true: Tensor
            Actually observed values. Should be greater than or equal to 0,
            but this will not be checked for, let alone enforced.

        Returns
        -------
        Tensor
            Proportional to the negative log-likelihood of a Tweedie
            distribution for 1 < `p` < 2.

        """
        p_ = p.clamp(1.0 + self.eps, 2.0 - self.eps)
        mu_ = mu + self.eps

        negll = - 2.0 * (
            y_true * mu_.pow(1.0 - p_) / (1.0 - p_) -
            y_true.pow(2.0 - p_) / (1.0 - p_) / (2.0 - p_) -
            mu_.pow(2.0 - p_) / (2.0 - p_)
        )
        return self._reduce(negll)


class BetaBernoulliLoss(_BaseLoss):
    """Special case of the Beta-Binomial negative log-likelihood for 1 trial.

    Use to make probabilistic predictions for binary classification, where
    the model outputs the alpha and beta coefficients of the Beta-distributed
    success probability instead of a point estimate of the success probability.

    Parameters
    ----------
    reduction: string, optional
        One of "mean", "sum" or "none". Defaults to "mean". Whether and, if so,
        how to aggregate the tensor resulting from evaluating the point-wise
        loss function on the input. Use the ``Reduction`` enum to avoid typos.
    eps: float, optional
        Many loss functions require input and/or target tensors to be bound
        by some lower and/or upper value. It is the user's responsibility to
        ensure that they are. However, evaluating a loss function just at the
        interval boundary of its support might lead to numerical inaccuracies.
        To avoid these, it is often advisable to shift such values away from
        boundaries by a small value `eps`. Defaults to 1e-6.

    See Also
    --------
    Reduction

    """

    def __init__(self, reduction: str = 'mean', eps: float = 1e-6) -> None:
        super().__init__(reduction, eps)

    def forward(self, alpha: Tensor, beta: Tensor, y_true: Tensor) -> Tensor:
        """Forward pass for the Beta-Binomial loss function.

        Parameters
        ----------
        alpha: Tensor
            Predicted values for the `alpha` parameter of the Beta-distributed
            success probability in binary classification. Should be greater
            than or equal to 0, but this will not be checked for, let alone
            enforced. However, `eps` is added to ensure numerical stability.
        beta: Tensor
            Predicted values for the `beta` parameter of the Beta-distributed
            success probability in binary classification. Should be greater
            than or equal to 0, but this will not be checked for, let alone
            enforced. However, `eps` is added to ensure numerical stability.
        y_true: Tensor
            Actually observed binary outcomes, encoded as 1.0 for success and
            0.0 for failure.

        Returns
        -------
        Tensor
            Negative log-likelihood of the Beta-Binomial distribution for the
            special case of one trial.

        """
        alpha_ = alpha + self.eps
        beta_ = beta + self.eps

        negll = -(
            pts.gammaln(alpha_ + y_true) +
            pts.gammaln(beta_ + (1.0 - y_true)) -
            pts.gammaln(alpha_) -
            pts.gammaln(beta_) -
            (alpha + beta + self.eps).log()
        )
        return self._reduce(negll)


class GammaLoss(_BaseLoss):
    """Negative log-likelihood of a scaled Gamma distribution.

    Potentially useful for strictly positive targets and skewed residuals.
    For convenience, the Gamma distribution is parameterized in terms of
    mean and standard deviation instead of its standard form in terms of
    shape and scale (or rate) parameters.

    Parameters
    ----------
    reduction: string, optional
        One of "mean", "sum" or "none". Defaults to "mean". Whether and, if so,
        how to aggregate the tensor resulting from evaluating the point-wise
        loss function on the input. Use the ``Reduction`` enum to avoid typos.
    eps: float, optional
        Many loss functions require input and/or target tensors to be bound
        by some lower and/or upper value. It is the user's responsibility to
        ensure that they are. However, evaluating a loss function just at the
        interval boundary of its support might lead to numerical inaccuracies.
        To avoid these, it is often advisable to shift such values away from
        boundaries by a small value `eps`. Defaults to 1e-6.

    See Also
    --------
    Reduction

    """

    def __init__(self, reduction: str = 'mean', eps: float = 1e-6) -> None:
        super().__init__(reduction, eps)

    def forward(self, mu: Tensor, sigma: Tensor, y_true: Tensor) -> Tensor:
        """Forward pass for the scaled Gamma loss function.

        Parameters
        ----------
        mu: Tensor
            Predicted mean values. Should be greater than or equal to 0, but
            this will not be checked for, let alone enforced. However, `eps`
            is added to ensure numerical stability.
        sigma: Tensor
            Predicted standard deviations. Should be greater than or equal
            to 0, but this will not be checked for, let alone enforced.
            However, `eps` is added to ensure numerical stability.
        y_true: Tensor
            Actually observed values. Should be greater than or equal
            to 0, but this will not be checked for, let alone enforced.
            However, `eps` is added to ensure numerical stability.

        Returns
        -------
        Tensor
            The negative log-likelihood of a scaled Gamma distribution.

        """
        mu_ = mu + self.eps
        sigma2 = sigma.pow(2.0) + self.eps
        ratio = mu_.pow(2.0) / sigma2
        y_true_ = y_true + self.eps
        scaled = y_true * mu / sigma2 + self.eps

        negll = -(
            ratio * scaled.log() -
            y_true_.log() -
            scaled -
            pts.gammaln(ratio)
        )
        return self._reduce(negll)


class StudentLoss(_BaseLoss):
    """Negative log-likelihood of a non-standardized Student's t distribution.

    Parameters
    ----------
    reduction: string, optional
        One of "mean", "sum" or "none". Defaults to "mean". Whether and, if so,
        how to aggregate the tensor resulting from evaluating the point-wise
        loss function on the input. Use the ``Reduction`` enum to avoid typos.
    eps: float, optional
        Many loss functions require input and/or target tensors to be bound
        by some lower and/or upper value. It is the user's responsibility to
        ensure that they are. However, evaluating a loss function just at the
        interval boundary of its support might lead to numerical inaccuracies.
        To avoid these, it is often advisable to shift such values away from
        boundaries by a small value `eps`. Defaults to 1e-6.

    See Also
    --------
    Reduction

    """

    def __init__(self, reduction: str = 'mean', eps: float = 1e-6) -> None:
        super().__init__(reduction, eps)
        self.register_buffer('const', -0.5 * pt.tensor(math.pi).log(), False)

    def forward(
            self,
            df: Tensor,
            loc: Tensor,
            scale: Tensor,
            y_true: Tensor
    ) -> Tensor:
        """Forward pass for the Student's t loss function.

        Parameters
        ----------
        df: tensor
            Predicted degrees of freedom. Should be greater than zero, but this
            will not be checked for, let alone enforced. However, `eps` is
            added for numerical stability.
        loc: Tensor
            Predicted mean values.
        scale: Tensor
            Predicted scales. Should be greater than or equal to 0, but this
            will not be checked for, let alone enforced. However, `eps` is
            added to ensure numerical stability.
        y_true:
            Actually observed values.

        Returns
        -------
        Tensor
            Negative log-likelihood of a Student's t distribution.

        """
        half_nu = 0.5 * df + self.eps
        half_nu1p = 0.5 * (df + 1.0)
        tau2nu = scale.pow(2.0) * df + self.eps

        negll = -(
            self.const +
            pts.gammaln(half_nu1p) -
            pts.gammaln(half_nu) -
            0.5 * tau2nu.log() -
            half_nu1p * ((y_true - loc).pow(2.0) / tau2nu).log1p()
        )
        return self._reduce(negll)


class NegativeBinomialLoss(_BaseLoss):
    """Negative log-likelihood of a Negative-Binomial distribution.

    Potentially useful for counts data where more flexibility than provided
    by a Poisson loss is desired because the data might be over-dispersed.
    For convenience, the Negative-Binomial distribution is parameterized in
    terms of mean and standard deviation instead of its standard form.

    Parameters
    ----------
    reduction: string, optional
        One of "mean", "sum" or "none". Defaults to "mean". Whether and, if so,
        how to aggregate the tensor resulting from evaluating the point-wise
        loss function on the input. Use the ``Reduction`` enum to avoid typos.
    eps: float, optional
        Many loss functions require input and/or target tensors to be bound
        by some lower and/or upper value. It is the user's responsibility to
        ensure that they are. However, evaluating a loss function just at the
        interval boundary of its support might lead to numerical inaccuracies.
        To avoid these, it is often advisable to shift such values away from
        boundaries by a small value `eps`. Defaults to 1e-6.

    Notes
    -----
    The present parameterization only makes sense if the variance is strictly
    greater than the mean. This is best taken into account already on the
    model side, e.g., by forcing the output for `sigma` to be a that of `mu`
    multiplied by one plus some (learnable) fraction of `mu`, as described
    by D. Salinas `et al.` in their DeepAR paper. [1]_

    See Also
    --------
    Reduction
    swak.pt.misc.NegativeBinomialFinalizer
    swak.pt.dists.MuSigmaNegativeBinomial

    References
    ----------
    .. [1] D. Salinas, V. Flunkert, and J. Gasthaus, and T. Pfister, `DeepAR:
           Probabilistic Forecasting with Autoregressive Recurrent Networks`,
           `arXiv:1704.04110v3 <https://arxiv.org/pdf/1704.04110>`__ (2019).

    """

    def __init__(self, reduction: str = 'mean', eps: float = 1e-6) -> None:
        super().__init__(reduction, eps)


    def forward(self, mu: Tensor, sigma: Tensor, y_true: Tensor) -> Tensor:
        """Forward pass for the Negative-Binomial loss function.

        Parameters
        ----------
        mu: Tensor
            Predicted mean values. Should be greater than or equal to 0, but
            this will not be checked for, let alone enforced. However, `eps`
            is added to ensure numerical stability.
        sigma: Tensor
            Predicted standard deviations. Should be greater than or equal
            to the square root of `mu`, but this will not be checked for,
            let alone enforced. However, `eps` is added to ensure numerical
            stability.
        y_true: Tensor
            Actually observed values. Should be greater than or equal
            to 0, but this will not be checked for, let alone enforced.

        Returns
        -------
        Tensor
            The negative log-likelihood of a Negative-Binomial distribution.

        """
        sigma2 = sigma.pow(2.0) + 2 * self.eps
        scaled = mu / sigma2 + self.eps
        ratio = mu.pow(2.0) / (sigma2 - mu) + self.eps

        negll = -(
            pts.gammaln(y_true + ratio) +
            y_true * (1.0 - scaled).log() +
            ratio * scaled.log() -
            pts.gammaln(y_true + 1.0) -
            pts.gammaln(ratio)
        )
        return self._reduce(negll)



from typing import Any
from torch.nn.modules.module import _IncompatibleKeys
from ..types import Optimizer
from ...misc import ArgRepr

__all__ = [
    'NoSchedule'
]


class NoSchedule(ArgRepr):
    """Mock learning-rate scheduler that does not actually do anything.

    This class is just a placeholder for an actual PyTorch learning-rate
    scheduler. Any additional (keyword) arguments provided at instantiation
    are ignored.

    Parameters
    ----------
    optimizer: Optimizer
        A PyTorch optimizer instance.

    """

    def __init__(
            self,
            optimizer: Optimizer,
            *_: Any,
            **__: Any
    ) -> None:
        super().__init__(optimizer)
        self.optimizer = optimizer

    def step(self, *args: Any, **kwargs: Any) -> None:
        """Does not do anything, certainly not touching the learning rate.

        Positional and/or keyword arguments can be provided, but are ignored.

        """

    def get_last_lr(self) -> list[float]:
        """Returns the default learning rate of the optimizer."""
        return [self.optimizer.defaults['lr']]

    def state_dict(self) -> dict[str, Any]:
        """Because there is no state return an empty, ordered dictionary."""
        return {}

    def load_state_dict(self, *_: Any, **__: Any) -> _IncompatibleKeys:
        """Do nothing, return the object thing as other methods of that type"""
        return _IncompatibleKeys([], [])



from typing import Any
from collections.abc import Callable


class Maybe[T]:
    """Type annotation to allow for ``None`` values.

    Initialize this class with a built-in type, class, or other callable that
    will cast a given value to the desired type if it is not ``None``. Then
    calling the callable instance on that value will return the desired type
    if the value is not ``None`` and ``None`` if it is.

    Parameters
    ----------
    cast: callable
        Casts value to desired type if value is not ``None``.

    Notes
    -----
    Upon instantiation, the generic class can be type-annotated with the
    return type of `cast`.

    """

    def __init__(self, cast: type[T] | Callable[[Any], T]) -> None:
        self.cast = cast

    def __call__(self, obj: Any) -> T:
        """Casts value to specified type if value is not ``None``.

        Parameters
        ----------
        obj
            Object to type-cast to, or ``None``.

        Returns
        -------
        object
            Type-cast object or ``None``.

        """
        obj_is_none_str = obj in ('null', 'None')
        return None if obj is None or obj_is_none_str else self.cast(obj)



from typing import Any, Literal
from enum import StrEnum
import pandas_gbq as pq
from pandas import DataFrame
from .exceptions import GbqError
from ...misc import ArgRepr


class IfExists(StrEnum):
    """Specify what to do if the BigQuery table to write to already exists."""
    FAIL = 'fail'
    REPLACE = 'replace'
    APPEND = 'append'


class DataFrame2Gbq(ArgRepr):
    """Partial of the ``to_gbq`` function in the``pandas_gbq`` package.

    As such, it may not be suitable for uploading large amounts of data (see
    its `documentation <https://googleapis.dev/python/pandas-gbq/latest/>`__
    for details).

    Parameters
    ----------
    project: str
        The project to bill for the upload.
    dataset: str
        The id of the dataset where the destination `table` resides.
    table: str, optional
        The name of the table to load data into (excluding the `dataset`) or
        the prefix to it. Defaults to an empty string but will be appended by
        the table name (or suffix) given on calling instances.
    location: str, optional
        The physical datacenter location to load the data onto. See the
        Google Cloud Platform `documentation <https://cloud.google.com/
        bigquery/docs/locations>`__ for options. If not given, non-existing
        tables will be created in the default location of the `dataset`.
        Defaults to an empty string.
    if_exists: str, optional
        What to do if the destination table already exists. Can be one of
        "fail", "replace" or "append". Defaults to "fail".  Use the
        ``IfExists`` enum to specify explicitly.
    chunksize: int, optional
        Number of dataframe rows to be inserted in each chunk.
        Defaults to ``None``, resulting in the entire dataframe to be
        inserted at once.
    **kwargs
        Additional keyword arguments passed on to the ``to_gbq`` call.

    See Also
    --------
    IfExists

    """

    def __init__(
            self,
            project: str,
            dataset: str,
            table: str = '',
            location: str = '',
            if_exists: Literal['fail', 'replace', 'append'] = 'fail',
            chunksize: int | None = None,
            **kwargs: Any
    ) -> None:
        self.project = project.strip(' ./')
        self.dataset = dataset.strip(' ./')
        self.table = table.strip(' ./')
        self.location = location.strip().lower()
        self.if_exists = if_exists.strip().lower()
        self.chunksize = chunksize
        self.kwargs = kwargs
        super().__init__(
            self.project,
            self.dataset,
            self.table,
            self.location,
            self.if_exists,
            self.chunksize,
            **kwargs
        )

    def __call__(self, df: DataFrame, table: str = '') -> tuple[()]:
        """Write a pandas DataFrame to a Google BigQuery table.

        Parameters
        ----------
        df: DataFrame
            The dataframe to upload to a Google BigQuery table.
        table: str, optional
            The name (or suffix) of the table to load data into (excluding the
            `dataset`), to be appended to the `table` given at instantiation.
            Defaults to an empty string.

        Returns
        -------
        tuple
            An empty tuple.

        Raises
        ------
        GbqError
            If no `table` was given, neither at instantiation nor when called.

        """
        table = self.table + table.strip(' ./')

        if not table:
            raise GbqError('You must provide a table name!')

        pq.to_gbq(
            df.reset_index(drop=True),
            project_id=self.project,
            destination_table=self.dataset + '.' + table,
            location=self.location or None,
            if_exists=self.if_exists,
            chunksize=self.chunksize,
            progress_bar=False,
            **self.kwargs
        )

        return ()



from typing import Any, Self, ParamSpec
from collections.abc import Iterator, Callable, Iterable
from functools import singledispatchmethod
from ..misc import IndentRepr
from .exceptions import ForkError

P = ParamSpec('P')
type Call = type | Callable[P, Any]


class Fork[**P, T](IndentRepr):
    """Call any number of callables with the same argument(s).

    Generic type annotation of instances is recommended. Provide a list of
    one or more input types that all callables take, followed by a ``tuple``
    specifying the concatenation of the return types of all callables, ignoring
    empty tuples. If only a single object remains, the type of that object
    should be annotated.

    Parameters
    ----------
    call: callable or iterable of callables, optional
        One callable or an iterator of callables to all call with the same
        arguments. Defaults to an empty tuple.
    *calls: callable
        Additional callables to call with the same argument(s).

    Raises
    ------
    ForkError
        If (any of) `call` or any of `calls` are not, in fact, callable.

    """

    def __init__(self, call: Call | Iterable[Call] = (), *calls: Call) -> None:
        self.calls = self.__valid(call) + self.__valid(calls)
        super().__init__(self.calls)

    def __iter__(self) -> Iterator[Call]:
        # We could also iterate over instances of self ...
        return self.calls.__iter__()

    def __len__(self) -> int:
        return self.calls.__len__()

    def __bool__(self) -> bool:
        return bool(self.calls)

    def __contains__(self, call: Call) -> bool:
        return call in self.calls

    def __reversed__(self) -> Self:
        return self.__class__(*reversed(self.calls))

    @singledispatchmethod
    def __getitem__(self, index: int) -> Call:
        # We could also return instances of self ...
        return self.calls[index]

    @__getitem__.register
    def _(self, index: slice) -> Self:
        return self.__class__(*self.calls[index])

    def __eq__(self, other: Self) -> bool:
        if isinstance(other, self.__class__):
            return self.calls == other.calls
        return NotImplemented

    def __ne__(self, other: Self) -> bool:
        if isinstance(other, self.__class__):
            return self.calls != other.calls
        return NotImplemented

    def __add__(self, other: Call | Iterable[Call] | Self) -> Self:
        if isinstance(other, self.__class__):
            return self.__class__(self.calls, *other.calls)
        try:
            return self.__class__(self.calls, *self.__valid(other))
        except ForkError:
            return NotImplemented

    def __radd__(self, other: Call | Iterable[Call] | Self) -> Self:
        if isinstance(other, self.__class__):
            return self.__class__(other.calls, *self.calls)
        try:
            return self.__class__(self.__valid(other), *self.calls)
        except ForkError:
            return NotImplemented

    def __call__(self, *args: P.args) -> T:
        """Call all specified `calls` with the same argument(s).

        Parameters
        ----------
        *args
            Arguments to call all `calls` with.

        Returns
        -------
        tuple or object
            Concatenation of all return values of all `calls` in order. If only
            one of the `calls` returns something other than an empty tuple,
            that object is returned.

        Raises
        ------
        ForkError
            When one of the `calls` raises an exception.

        """
        results = []
        for i, call in enumerate(self.calls):
            try:
                result = call(*args)
            except Exception as error:
                msg = '\n{} executing\n{}\nin fork {} of\n{}\n{}'
                err_cls = error.__class__.__name__
                name = self._name(call)
                fmt = msg.format(err_cls, name, i, self, error)
                raise ForkError(fmt) from error
            else:
                if isinstance(result, tuple):
                    results.extend(result)
                else:
                    results.append(result)
        return results[0] if len(results) == 1 else tuple(results)

    @staticmethod
    def __valid(calls: Call | Iterable[Call]) -> tuple[Call, ...]:
        """Ensure that the argument is indeed an iterable of callables."""
        if callable(calls):
            return calls,
        iterable = True
        all_callable = False
        try:
            all_callable = all(callable(call) for call in calls)
        except TypeError:
            iterable = False
        if iterable and all_callable:
            return tuple(calls)
        raise ForkError('All branches in the fork must be callable!')



import string
from typing import Any
from collections.abc import Mapping

type M = Mapping[str, Any] | None


class _Template(string.Template):
    """String template accepting also dots in the substitution keys.

    This is a drop-in replacement for the python standard library's
    ``string.Template`` with the only difference that, for braced substitution
    keys (e.g., "My name is ${person.name}.") dots are allowed as well.

    """

    braceidpattern = '(?a:[_a-z][_a-z0-9]*(\\.[_a-z][_a-z0-9]*)*)'


class TemplateRenderer:
    """Substitute bash-style "${key}" placeholders in a template string.

    This is a light wrapper around the standard library's ``string.Template``,
    calling its method ``safe_substitute`` upon instantiation and its method
    ``substitute`` when calling the (callable) instance. At first, not all keys
    need to be provided, but when the object is called and the substitution
    is finalized, all keys must have been provided.

    Parameters
    ----------
    template: str
        String with bash-style "${key}" placeholders to substitute.
    mapping: dict, optional
        Dictionary-like mapping with the placeholders to substitute as keys
        and the values to substitute as values. Defaults to ``None``.
    **kwargs
        Keyword arguments replace respective placeholders with whatever value
        is passed, overwriting values in `mapping`.

    Notes
    -----
    Instances evaluate to ``True`` if all keys were provided at instantiation
    and to ``False`` if there are still keys missing.

    """

    def __init__(
            self,
            template: str,
            mapping: M = None,
            **kwargs: Any
    ) -> None:
        mapping = {} if mapping is None else mapping
        # We could also delay this until instances are called
        self.template = _Template(template).safe_substitute(mapping, **kwargs)

    def __repr__(self) -> str:
        cls = self.__class__.__name__
        # Since the string template can be very large, we only
        # display up to 30 characters of its first line.
        head = self.template.splitlines()[0].strip()
        suffix = ' ...' if len(head) > 30 else ''
        return f"{cls}('{head[:30].strip() + suffix}')"

    def __str__(self) -> str:
        return self.template

    def __bool__(self) -> bool:
        return not bool(self.identifiers)

    @property
    def identifiers(self) -> list[str]:
        """List of identifiers that still need to be interpolated."""
        return _Template(self.template).get_identifiers()

    def __call__(self, mapping: M = None, **kwargs: Any) -> str:
        """Substitute bash-style "${key}" placeholders in the wrapped template.

        Parameters
        ----------
        mapping: dict, optional
            Dictionary-like mapping with the placeholders to substitute as keys
            and the values to substitute as values. Defaults to ``None``.
        **kwargs
            Keyword arguments replace respective placeholders with whatever
            value is passed, overwriting values in `mapping`.

        Returns
        -------
        str
            String with all placeholders substituted.

        Raises
        ------
        KeyError
            When a value for one or more placeholders is still missing.
        ValueError
            When a key is not a valid python identifier.

        """
        mapping = {} if mapping is None else mapping
        return _Template(self.template).substitute(mapping, **kwargs)


class FormFiller:
    """Substitute bash-style "${key}" placeholders in a template string.

    This is a light wrapper around the standard library's ``string.Template``,
    first caching a dictionary and optional keywords on instantiation, and
    then calling its method ``substitute`` with these cached values when
    calling the (callable) object with the template string.

    Parameters
    ----------
    mapping: dict, optional
        Dictionary-like mapping with the placeholders to substitute as keys
        and the values to substitute as values. Defaults to ``None``.
    **kwargs
        Keyword arguments replace respective placeholders with whatever value
        is passed, overwriting respective values in `mapping`.

    Notes
    -----
    Instances evaluate to ``False`` if no keys were provided at instantiation
    and to ``True`` if there are any keys cached.

    """

    def __init__(self, mapping: M = None, **kwargs: Any) -> None:
        self.mapping = kwargs if mapping is None else {**mapping, **kwargs}

    def __repr__(self) -> str:
        cls = self.__class__.__name__
        # Since the mapping can be very large, we only
        # display up to its first 30 characters.
        head = str(self.mapping).splitlines()[0].strip()
        suffix = ' ...}' if len(head) > 30 else ''
        return f'{cls}({head[:30] + suffix})'

    def __bool__(self) -> bool:
        return bool(self.mapping)

    def __call__(self, template: str, mapping: M = None, **kwargs: Any) -> str:
        """Substitute bash-style "${key}" placeholders in a string template.

        Parameters
        ----------
        template: str
            String with bash-style "${key}" placeholders to substitute.
        mapping: dict, optional
            Dictionary-like mapping with the placeholders to substitute as keys
            and the values to substitute as values. Defaults to ``None``.
        **kwargs
            Keyword arguments replace respective placeholders with whatever
            value is passed, overwriting the cached `mapping`.

        Returns
        -------
        str
            String with all placeholders substituted.

        Raises
        ------
        KeyError
            When a value for one or more placeholders is still missing.
        ValueError
            When a key is not a valid python identifier.

        """
        mapping = kwargs if mapping is None else {**mapping, **kwargs}
        return _Template(template).substitute(self.mapping, **mapping)



from typing import Self
import torch as pt
from ...types import Tensor, Module


class ConstantSumMixer(Module):
    """Combine stacked feature vectors by simply adding them.

    The sum is then "normed" through dividing by the number of features.

    Parameters
    ----------
    n_features: int
        The number of features to combine. Must be equal to the size of the
        next-to-last dimension of the input tensor.

    """

    def __init__(self, n_features: int):
        super().__init__()
        self.n_features = n_features
        self.register_buffer('coeffs', pt.tensor(1.0 / n_features))

    def importance(self, inp: Tensor) -> Tensor:
        """Constant feature weights in the normed sum over all features.

        Parameters
        ----------
        inp: Tensor
            Feature vectors stacked into a tensor of at least 2 dimensions.
            The size of the next-to-last last dimension is expected to match
            the `n_features` provided at instantiation. The last dimension is
            expected to contain the features vectors themselves.

        Returns
        -------
        Tensor
            The output tensor has one fewer dimensions than the input with the
            last dimension being dropped.

        """
        return self.coeffs.expand(*inp.shape[:-1])

    def forward(self, inp: Tensor) -> Tensor:
        """Add stacked feature vectors with constant and equal weights.

        Parameters
        ----------
        inp: Tensor
            Feature vectors stacked into a tensor of at least 2 dimensions.
            The size of the next-to-last last dimension is expected to match
            the `n_features` provided at instantiation. The last dimension is
            expected to contain the features vectors themselves.

        Returns
        -------
        Tensor
            The output tensor has one fewer dimensions than the input.
            The next-to-last dimension is dropped and the last dimension now
            contains the (normed) sum of all feature vectors.

        """
        return (self.importance(inp).unsqueeze(dim=-2) @ inp).squeeze(dim=-2)

    def reset_parameters(self) -> None:
        """Does nothing because there are no internal parameters to reset."""

    def new(self, n_features: int | None = None) -> Self:
        """Return a fresh instance with the same or updated parameters.

        Parameters
        ----------
        n_features: int, optional
            The number of features to combine. Must be equal to the size of the
            next-to-last dimension of the input tensor. Overwrites `n_features`
            of the current instance if given. Defaults to ``None``.

        Returns
        -------
        ConstantSumMixer
            A fresh, new instance of itself.

        """
        return self.__class__(
            self.n_features if n_features is None else n_features
        )



from enum import StrEnum


class NotFound(StrEnum):
    """Enum to direct read/load behaviour in case of missing files.

    See Also
    --------
    TomlReader, YamlReader

    """
    IGNORE = 'ignore'
    WARN = 'warn'
    RAISE = 'raise'



"""Special field types for type-annotating class variables in a JsonObject."""

from .maybe import Maybe
from .flexidate import FlexiDate
from .flexitime import FlexiTime
from .custom import CustomField

__all__ = [
    'Maybe',
    'FlexiDate',
    'FlexiTime',
    'CustomField'
]



from typing import Any
from collections.abc import Callable, Iterable
from concurrent.futures import ThreadPoolExecutor
from ...misc import ArgRepr
from ..exceptions import MapError


class ThreadMap[**P, S, T](ArgRepr):
    """Partial of ``concurrent.futures.ThreadPoolExecutor.map``.

    Upon subclassing and/or instantiation, type annotation with a list of the
    argument type(s) of `transform`, the return type of `call`, and the return
    type of `wrapper` is recommended.

    Parameters
    ----------
    transform: callable
        Transforms element(s) of the input iterable(s).
    wrapper: type or callable, optional
        If not given, an attempt will be made to return the type of the first
        iterable the callable instance is being called with (by calling its
        class with a list of the mapped elements). If explicitly given,
        `wrapper` will be called with a list of mapped elements. Consequently,
        the return type will be the (return) type of `wrapper`.
    max_workers: int, optional
        Maximum number of worker threads used in the pool to execute
        `transform` asynchronously. Will be forwarded to the constructor of
        ``ThreadPoolExecutor``. Defaults to 16.
    thread_name_prefix: str, optional
        Will be forwarded to the constructor of ``ThreadPoolExecutor``.
        Defaults to an empty string.
    initializer: callable, optional
        Called at the start of each worker thread. Will be forwarded to the
        constructor of ``ThreadPoolExecutor``. Defaults to ``None``.
    initargs: tuple, optional
        Arguments passed to the initializer. Will be forwarded to the
        constructor of ``ThreadPoolExecutor``. Defaults to an empty tuple.
    timeout: int or float, optional
        Maximum time (in seconds) to wait for results to be available. Defaults
        to ``None``, which means there is no limit for the time to wait. Will
        be forwarded to the ``map`` method of the ``ThreadPoolExecutor``.

    Notes
    -----
    In contrast to calling the ``map`` method of a ``ThreadPoolExecutor``
    directly, which returns a generator object, the mapped iterable is fully
    manifested first and only then wrapped.

    See Also
    --------
    concurrent.futures.ThreadPoolExecutor

    """

    def __init__(
            self,
            transform: type[S] | Callable[P, S],
            wrapper: type[T] | Callable[[list[S]], T] | None = None,
            max_workers: int = 16,
            thread_name_prefix: str = '',
            initializer: Callable[..., Any] | None = None,
            initargs: tuple[Any, ...] = (),
            timeout: float | None = None
    ) -> None:
        super().__init__(
            transform,
            wrapper,
            max_workers,
            thread_name_prefix,
            initializer,
            initargs,
            timeout
        )
        self.transform = transform
        self.wrapper = wrapper
        self.max_workers = max_workers
        self.thread_name_prefix = thread_name_prefix
        self.initializer = initializer
        self.initargs = initargs
        self.timeout = timeout

    def __call__(self, iterable: Iterable, *iterables: Iterable) -> T:
        """Concurrently transform the element(s) of the given iterable(s).

        Parameters
        ----------
        iterable: Iterable
            An iterable of elements to transform.
        *iterables: Iterable
            If given, the cached `transform` is called with the corresponding
            elements of `ìterable` and all `iterables` as arguments.

        Returns
        -------
        Sequence
            Same type as `iterable` if `wrapper` was not specified on
            instantiation or the (return) type of `wrapper`. Note that, as with
            python's builtin ``map`` function, the length of the output
            sequence is limited by the shortest of the input iterables.

        Raises
        ------
        MapError
            If calling the ``ThreadPoolExecutor``'s ``map`` method raises an
            exception or if wrapping the results leads to an exception.

        """
        with ThreadPoolExecutor(
            self.max_workers,
            self.thread_name_prefix,
            self.initializer,
            self.initargs
        ) as pool:
            mapped = pool.map(
                self.transform,
                iterable,
                *iterables,
                timeout=self.timeout
            )
            try:
                mapped = list(mapped)
            except Exception as error:
                msg = ('Error calling\n{}\non one or more '
                       'element(s) of the iterable(s)!\n{}:\n{}')
                name = self._name(self.transform)
                err_cls = error.__class__.__name__
                fmt = msg.format(name, err_cls, error)
                raise MapError(fmt) from error
        wrap = iterable.__class__ if self.wrapper is None else self.wrapper
        try:
            wrapped = wrap(mapped)
        except Exception as error:
            msg = '\n{} calling wrapper\n{}\non map results:\n{}'
            name = self._name(wrap)
            err_cls = error.__class__.__name__
            raise MapError(msg.format(err_cls, name, error)) from error
        return wrapped



from enum import StrEnum
from typing import Any, Literal
from google.cloud.storage import Client, Bucket
from google.api_core.retry import Retry
from ...misc import ArgRepr

type StorageType = Literal['STANDARD', 'NEARLINE', 'COLDLINE', 'ARCHIVE']


class Storage(StrEnum):
    """Specify storage class for blobs in a Google Cloud Storage bucket."""
    STANDARD = 'STANDARD'
    NEARLINE = 'NEARLINE'
    COLDLINE = 'COLDLINE'
    ARCHIVE = 'ARCHIVE'


class GcsBucket(ArgRepr):
    """Create a new bucket on Google Cloud Storage.

    Parameters
    ----------
    project: str
        The project to create the bucket in.
    bucket: str
        The name of the bucket to create.
    location: str
        The physical datacenter location to create the bucket in. See the
        Google Cloud Platform `documentation <https://cloud.google.com/storage/
        docs/locations>`__ for options.
    blob_expire_days: int, optional
        Defaults to ``None``. If sets, blobs older than the specified number of
        days will be automatically deleted.
    labels: dict, optional
        Any number of string-valued labels of the bucket. Defaults to none.
    user_project: str, optional
        The project billed for interacting with the bucket. Defaults to the
        `project`
    storage_class: str, optional
        Defaults storage class for blobs in this bucket. Defaults to ``None``,
        which results in "STANDARD".  Use the ``Storage`` enum to specify
        explicitly.
    requester_pays: bool, optional
        Whether the requester will be billed for interacting with the bucket.
        Defaults to ``False``, which means that the (`user_`)`project` will be
        billed.
    **kwargs
        Additional keyword arguments are passed to the constructor of the
        Google Storage ``Client`` (see `documentation <https://cloud.google.
        com/python/docs/reference/storage/latest/google.cloud.storage.
        client.Client#parameters>`__ for options).

    Notes
    -----
    There are a lot more options to set, which have been deliberately omitted
    because of the complexity involved.

    See Also
    --------
    Storage

    """

    def __init__(
            self,
            project: str,
            bucket: str,
            location: str,
            blob_expire_days: int | None = None,
            labels: dict[str, str] | None = None,
            user_project: str | None = None,
            storage_class: StorageType | None = None,
            requester_pays: bool = False,
            **kwargs: Any
    ) -> None:
        self.project = project.strip(' /.')
        self.bucket = bucket.strip(' /.')
        self.location = location.strip().upper()
        self.blob_expire_days = blob_expire_days
        self.labels = {} if labels is None else labels
        if user_project is None:
            self.user_project = self.project
        else:
            self.user_project = user_project.strip(' /.')
        self.storage_class = storage_class
        self.requester_pays = requester_pays
        self.kwargs = kwargs
        super().__init__(
            self.project,
            self.bucket,
            self.location,
            blob_expire_days=self.blob_expire_days,
            labels=self.labels,
            user_project=self.user_project,
            storage_class=self.storage_class,
            requester_pays=self.requester_pays,
            **kwargs
        )

    def __call__(
            self,
            exists_ok: bool = True,
            retry: Retry | None = None,
            timeout: float | tuple[float, float] | None = None,
    ) -> tuple[Bucket, bool]:
        """Create a new bucket on Google Cloud Storage.

        Parameters
        ----------
        exists_ok: bool, optional
            Whether to raise a ``Conflict`` exception if the targeted bucket
            already exists or not. Defaults to ``True``.
        retry: Retry, optional
            Retry policy for the request. Defaults to ``None``, which disables
            retries. See the Google Cloud Platform `guide
            <https://cloud.google.com/python/docs/reference/storage/1.39.0/
            retry_timeout#configuring-retries>`__ and `reference
            <https://googleapis.dev/python/google-api-core/latest/retry.html>`__
            for options.
        timeout: float, optional
            The number of seconds to wait for the HTTP response to the API call
            before using `retry` or a tuple with separate values for connection
            and request timeouts. Defaults to ``None``, meaning wait forever.

        Raises
        ------
        Conflict
            If `exists_ok` is set to ``False`` and the dataset already exists.

        Returns
        -------
        Bucket
            The existing or newly created bucket. If existing, then the bucket
            is returned unchanged, that is, none of the specified options are
            applied.
        bool
            ``True`` if the requested bucket is newly created and ``False``
            if an existing bucket is returned.

        """
        client = Client(self.project, **self.kwargs)
        bucket = Bucket(client, self.bucket, self.user_project)

        bucket.requester_pays = self.requester_pays
        bucket.storage_class = self.storage_class
        bucket.labels = self.labels
        if self.blob_expire_days:
            bucket.add_lifecycle_delete_rule(age=self.blob_expire_days)

        if bucket.exists() and exists_ok:
            existing = client.get_bucket(bucket, retry=retry, timeout=timeout)
            return existing, False

        bucket.create(
            client,
            self.project,
            self.location,
            retry=retry,
            timeout=timeout
        )
        return bucket, True



"""Building blocks and tools to construct, train, and manage PyTorch models.

Cast your data into tensors, embed your features, combine them (optionally
extracting feature importance), and pass them through repeated residual layers
of flexible architecture. Custom loss functions and re-parameterized
distributions allow you to make probabilistic predictions. The training loop
is conveniently abstracted away, allowing for (ready-to-use or custom-made)
callbacks and checkpoints. Finally trained models can be saved and (re)loaded.

Notes
-----
The tools in this subpackage are mostly geared towards tabular and sequence
data. Image processing is not currently the main focus.

"""

try:
    import torch as pt
except ModuleNotFoundError:
    msg = 'Install the "torch" package to unlock the PyTorch utilities!'
    raise ImportError(msg)

__all__ = ['device']

device = pt.device('cuda') if pt.cuda.is_available() else pt.device('cpu')
pt.set_default_device(device)
pt.set_float32_matmul_precision('medium')

# ToDo: Implement tools for packed & padded sequences
